/*
 * C
 *
 * Copyright 2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

// -----------------------------------------------------------------------------
// Includes
// -----------------------------------------------------------------------------

// OS includes
#include "FreeRTOS.h"
#include "timers.h"

// Driver includes
#include "board.h"
#include "fsl_debug_console.h"
#include "fsl_usart_dma.h"
#include "fsl_flexcomm.h"

#include "LLFEATURE_INPUTSTREAM.h"
#include "power_manager.h"

// -----------------------------------------------------------------------------
// Macros and defines
// -----------------------------------------------------------------------------

/*
 * @brief Buffer where write the feature (.fo) file.
 */
extern uint32_t FEATURE_LOADER_RAM_BUFFER_START;

#define USART_DBG         USART0
#define USART_DBG_IRQ     FLEXCOMM0_IRQn
#define USART_DMA         DMA0
#define USART_DMA_CHANNEL 0

// ID for the FreeRTOS Timer
#define MONITORTIMER_ID	20
#define MONITORTIMER_PERIOD_MS	1000

#define ELF_FILE_START 0x7f454c46

// -----------------------------------------------------------------------------
// Global variables
// -----------------------------------------------------------------------------

static uint32_t roll_buffer;

static dma_handle_t g_DMA_Handle;

DMA_ALLOCATE_LINK_DESCRIPTORS(dmaDesc, 2);
static dma_descriptor_t *currentDesc, *nextDesc;

static uint8_t *buffer_ptr;
static uint32_t transfered_bytes;

static TimerHandle_t MonitorTimer;

static uint16_t transfer_count = 0xFFFF;


// -----------------------------------------------------------------------------
// Private functions prototypes
// -----------------------------------------------------------------------------

static void TransferMonitor(TimerHandle_t timer);
static void UartRxCallback(USART_Type *base, void *handle);
static void DMACallback(dma_handle_t *handle, void *param, bool transferDone, uint32_t tcds);
static void GetDMATransferConfig(dma_transfer_config_t *transferConfig);
static void DMATransferInit(void);
static void DMANextDescConfig(void);

// -----------------------------------------------------------------------------
// Private functions
// -----------------------------------------------------------------------------

/*
 * @brief Freertos TransferMonitor Timer callback
 * Checks that the DMA Channel transfer count has changed since
 * last call. Otherwise it will terminate the DMA transfer.
 */
static void TransferMonitor(TimerHandle_t timer) {

	// Check every second DMA transfer status
	if(transfer_count == DMA_GetRemainingBytes(USART_DMA, USART_DMA_CHANNEL))
	{
		DMA_AbortTransfer(&g_DMA_Handle);

		transfered_bytes += DMA_MAX_TRANSFER_COUNT - transfer_count;
		PRINTF("Transferred bytes: %d\n", transfered_bytes);
		PRINTF("Transfer is over\n");

		// fill first address with ELF tag (to unlock feature inputstream)
		// Except is the transfert was interrupted by a call to USART_Feature_Loader_Disable()
		if(USART_DBG->FIFOCFG & USART_FIFOCFG_DMARX_MASK)
		{
			*((uint32_t *)FEATURE_LOADER_RAM_BUFFER_START) = FEATURE_SIGNATURE_VALUE;
		}

		// Disable USART connection to DMA
		USART_EnableRxDMA(USART_DBG, false);


		// Stop monitor timer
		xTimerStop(MonitorTimer, 0);

		// Enable lowpower
		power_manager_enable_low_power(POWER_APPLOADER, LOW_POWER_AUTHORIZED);

		// Configure Next DMA transfer
		DMATransferInit();

		// Configure next DMA descriptor
		DMANextDescConfig();

		// Enable USART Interrupt to receive next feature file
		USART_EnableInterrupts(USART_DBG, USART_FIFOINTENCLR_RXLVL_MASK);
	}
	else
	{
		transfer_count = DMA_GetRemainingBytes(USART_DMA, USART_DMA_CHANNEL);
	}
}

// UART IRQ callback function
static void UartRxCallback(USART_Type *base, void *handle) {
	char ch;

	if ((USART_DBG->FIFOSTAT & USART_FIFOSTAT_RXNOTEMPTY_MASK) != 0U) {
		ch = (uint8_t) USART_DBG->FIFORD;
		roll_buffer = (roll_buffer << 8) | ch;

		// Compare with start of ELF file
		if (ELF_FILE_START == roll_buffer) {

			// Disable USART interrupt
			USART_DisableInterrupts(USART_DBG, USART_FIFOINTENCLR_RXLVL_MASK);

			// Connect DMA to USART
			USART_EnableRxDMA(USART_DBG, true);

			// Start MonitorTimer
			xTimerStart(MonitorTimer, 0);

			// Disable lowpower
			power_manager_enable_low_power(POWER_APPLOADER, LOW_POWER_FORBIDDEN);
		}
	}
}

// DMA callback function
static void DMACallback(dma_handle_t *handle, void *param, bool transferDone,
		uint32_t tcds) {
	// Rearm transfer_count after each received descriptor
	transfer_count = 0xFFFF;

	DMANextDescConfig();
	transfered_bytes += DMA_MAX_TRANSFER_COUNT;
	PRINTF("Transfered bytes: %d\n", transfered_bytes);
}

static void GetDMATransferConfig(dma_transfer_config_t *transferConfig) {

	transferConfig->srcAddr = (void*) &USART_DBG->FIFORD;
	transferConfig->dstAddr = buffer_ptr;

	transferConfig->nextDesc = (uint8_t*) nextDesc;
	transferConfig->xfercfg.reload = true;

	transferConfig->isPeriph = true;

	transferConfig->xfercfg.swtrig = false;
	transferConfig->xfercfg.clrtrig = false;
	transferConfig->xfercfg.intA = true;
	transferConfig->xfercfg.intB = false;
	transferConfig->xfercfg.byteWidth = 1;
	transferConfig->xfercfg.srcInc = 0;
	transferConfig->xfercfg.dstInc = 1;
	transferConfig->xfercfg.transferCount = DMA_MAX_TRANSFER_COUNT;

	transferConfig->xfercfg.valid = true;

}

/*
 * @brief Configure the DMA Channel
 */
static void DMATransferInit(void) {
	dma_transfer_config_t transferConfig;

	// Initialize global variables
	transfered_bytes = 0;
	buffer_ptr = (uint8_t*) (FEATURE_LOADER_RAM_BUFFER_START + 4);
	// Initialize DMA
	DMA_Init(USART_DMA);
	DMA_CreateHandle(&g_DMA_Handle, USART_DMA, USART_DMA_CHANNEL);
	DMA_EnableChannel(USART_DMA, USART_DMA_CHANNEL);

	DMA_SetCallback(&g_DMA_Handle, DMACallback, NULL);

	// Configure first tranfert
	GetDMATransferConfig(&transferConfig);

	DMA_SubmitTransfer(&g_DMA_Handle, &transferConfig);

	// Increment buffer_ptr for next descriptor
	buffer_ptr += DMA_MAX_TRANSFER_COUNT;

	DMA_StartTransfer(&g_DMA_Handle);

}

/*
 * @brief Configure the next DMA Descriptor
 */
static void DMANextDescConfig(void) {
	dma_transfer_config_t transferConfig;
	dma_descriptor_t *tempDesc;

	// Switch Descriptors
	tempDesc = nextDesc;
	nextDesc = currentDesc;
	currentDesc = tempDesc;

	// Configure descriptor that has previously been linked
	GetDMATransferConfig(&transferConfig);
	DMA_CreateDescriptor(currentDesc, &transferConfig.xfercfg,
			(void*) &USART0->FIFORD, buffer_ptr, nextDesc);

	// Increment buffer_ptr for next descriptor
	buffer_ptr += DMA_MAX_TRANSFER_COUNT;
}

// -----------------------------------------------------------------------------
// Public functions
// -----------------------------------------------------------------------------

// See the header file for the function documentation
void USART_Feature_Loader_Init(void) {

	// Initialize static variables
	roll_buffer = 0;

	currentDesc = &dmaDesc[0];
	nextDesc = &dmaDesc[1];

	// Configure USART irq to detect start of ELF file
	USART_EnableInterrupts(USART_DBG, USART_FIFOINTENCLR_RXLVL_MASK);
	FLEXCOMM_SetIRQHandler(USART_DBG,
			(flexcomm_irq_handler_t) UartRxCallback, NULL);
	// Set high priority for USART_IRQ
	NVIC_SetPriority(FLEXCOMM0_IRQn, 2);
	EnableIRQ(FLEXCOMM0_IRQn);

	// Configure Next DMA transfer
	DMATransferInit();

	// Configure next DMA descriptor
	DMANextDescConfig();

	MonitorTimer = xTimerCreate(
			NULL,						// timer name
			MONITORTIMER_PERIOD_MS / portTICK_PERIOD_MS,	// timer period in (FreeRTOS) ticks
			true,					// auto reload
			(void*) MONITORTIMER_ID,	// timer Id
			TransferMonitor	// callback function
	);

}

// See the header file for the function documentation
void USART_Feature_Loader_Disable(void) {
	// Disable USART interrupt
	USART_DisableInterrupts(USART_DBG, USART_FIFOINTENCLR_RXLVL_MASK);

	// Disable USART connection to DMA
	USART_EnableRxDMA(USART_DBG, false);
}

// See the header file for the function documentation
void USART_Feature_Loader_Enable(void) {
	// Enable USART interrupt
	USART_EnableInterrupts(USART_DBG, USART_FIFOINTENCLR_RXLVL_MASK);
}
