/**************************************************************************//**
 * Copyright 2018, Silicon Laboratories Inc.
 *
 * Copyright 2020-2021 MicroEJ Corp. This file has been modified by MicroEJ Corp.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *****************************************************************************/


#include <stdbool.h>
#include "sl_wfx_host_pin.h"
#include "FreeRTOS.h"

#include "FreeRTOS.h"
#include "semphr.h"
#include "task.h"
#include "board.h"

#include "fsl_gpio.h"
#include "fsl_spi.h"
#include "fsl_flexio_spi.h"


#define BOARD_FLEXIO_BASE FLEXIO0
#define FLEXIO_SPI_MOSI_PIN 13U
#define FLEXIO_SPI_MISO_PIN 14U
#define FLEXIO_SPI_SCK_PIN 15U
#define FLEXIO_SPI_PCS0_PIN 12U

/* Select flexio clock source */
#define FLEXIO_CLOCK_SRC (kFRO96M_to_FLEXIO)
/* Clock divider for flexio clock source */
#define FLEXIO_CLOCK_DIVIDER (8U)
#define FLEXIO_CLOCK_FREQUENCY CLOCK_GetFlexioClkFreq()
#define TRANSFER_SIZE 256U        /*! Transfer dataSize */
#define TRANSFER_BAUDRATE 500000U /*! Transfer baudrate - 500k */


extern TaskHandle_t busCommTaskHandle;


// These must be same ie either A or B.
#define SPI_WIRQ_INTERRUPT_ID       kGPIO_InterruptA

/*******************************************************************************
 * Prototypes
 ******************************************************************************/
/* SPI user callback */
void FLEXIO_SPI_MasterUserCallback(FLEXIO_SPI_Type *base,
                                   flexio_spi_master_handle_t *handle,
                                   status_t status,
                                   void *userData);

/*******************************************************************************
 * Variables
 ******************************************************************************/
uint8_t masterTxData[TRANSFER_SIZE] = {0U};

FLEXIO_SPI_Type spiDev;
flexio_spi_master_handle_t g_m_handle;

volatile bool isTransferCompleted = false;

/*******************************************************************************
 * Code
 ******************************************************************************/

void FLEXIO_SPI_MasterUserCallback(FLEXIO_SPI_Type *base,
                                   flexio_spi_master_handle_t *handle,
                                   status_t status,
                                   void *userData)
{
    if (status == kStatus_Success)
    {
        __NOP();
    }
}


/*******************************************************************************
 * Code
 ******************************************************************************/
/*!
 * @brief Interrupt service function of switch.
 */
void SPI_WIRQ_INTERRUPT_HANDLER(void)
{
  portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
   

  if ( GPIO_PortGetInterruptStatus(GPIO, SL_WFX_IRQ_PORT_SPI, SPI_WIRQ_INTERRUPT_ID) & 
       (1u << SL_WFX_IRQ_GPIO_SPI ) ) 
  {

    // Notify busCommTaskHandle 
    vTaskNotifyGiveFromISR( busCommTaskHandle, &xHigherPriorityTaskWoken );

    /* clear the interrupt status */
    GPIO_PinClearInterruptFlag(GPIO, 
                               SL_WFX_IRQ_PORT_SPI, 
                               SL_WFX_IRQ_GPIO_SPI, 
                               SPI_WIRQ_INTERRUPT_ID);

  	/* Switch tasks if necessary */
    if (xHigherPriorityTaskWoken != pdFALSE)
    {
      portEND_SWITCHING_ISR(xHigherPriorityTaskWoken);
    }


  }


/* Add for ARM errata 838869, affects Cortex-M4, Cortex-M4F Store immediate overlapping
  exception return operation might vector to incorrect interrupt */
#if defined __CORTEX_M && (__CORTEX_M == 4U)
    __DSB();
#endif
}

//SemaphoreHandle_t spiDMASemaphore;

// This function is copied from FLEXIO_SPI_MasterInit() in fsl_flexio_spi.c  and customized
// for wfx module.
/*!
 * brief Ungates the FlexIO clock, resets the FlexIO module, configures the FlexIO SPI master hardware,
 * and configures the FlexIO SPI with FlexIO SPI master configuration. The
 * configuration structure can be filled by the user, or be set with default values
 * by the FLEXIO_SPI_MasterGetDefaultConfig().
 *
 * note FlexIO SPI master only support CPOL = 0, which means clock inactive low.
 *
 * Example
   code
   FLEXIO_SPI_Type spiDev = {
   .flexioBase = FLEXIO,
   .SDOPinIndex = 0,
   .SDIPinIndex = 1,
   .SCKPinIndex = 2,
   .CSnPinIndex = 3,
   .shifterIndex = {0,1},
   .timerIndex = {0,1}
   };
   flexio_spi_master_config_t config = {
   .enableMaster = true,
   .enableInDoze = false,
   .enableInDebug = true,
   .enableFastAccess = false,
   .baudRate_Bps = 500000,
   .phase = kFLEXIO_SPI_ClockPhaseFirstEdge,
   .direction = kFLEXIO_SPI_MsbFirst,
   .dataMode = kFLEXIO_SPI_8BitMode
   };
   FLEXIO_SPI_MasterInit(&spiDev, &config, srcClock_Hz);
   endcode
 *
 * param base Pointer to the FLEXIO_SPI_Type structure.
 * param masterConfig Pointer to the flexio_spi_master_config_t structure.
 * param srcClock_Hz FlexIO source clock in Hz.
*/
static void FLEXIO_SPI_MasterInit_for_wfx(FLEXIO_SPI_Type *base, flexio_spi_master_config_t *masterConfig, uint32_t srcClock_Hz)
{
    // These are the modifications to FLEXIO_SPI_MasterInit() in fsl_flexio_spi.c 
    flexio_pin_polarity_t pinPolarity = kFLEXIO_PinActiveHigh;
    flexio_timer_trigger_polarity_t triggerPolarity  = kFLEXIO_TimerTriggerPolarityActiveLow;

    assert(base);
    assert(masterConfig);

    flexio_shifter_config_t shifterConfig;
    flexio_timer_config_t timerConfig;
    uint32_t ctrlReg  = 0;
    uint16_t timerDiv = 0;
    uint16_t timerCmp = 0;

    /* Clear the shifterConfig & timerConfig struct. */
    memset(&shifterConfig, 0, sizeof(shifterConfig));
    memset(&timerConfig, 0, sizeof(timerConfig));

#if !(defined(FSL_SDK_DISABLE_DRIVER_CLOCK_CONTROL) && FSL_SDK_DISABLE_DRIVER_CLOCK_CONTROL)
    /* Ungate flexio clock. */
    CLOCK_EnableClock(s_flexioClocks[FLEXIO_GetInstance(base->flexioBase)]);
#endif /* FSL_SDK_DISABLE_DRIVER_CLOCK_CONTROL */

    /* Configure FLEXIO SPI Master */
    ctrlReg = base->flexioBase->CTRL;
    ctrlReg &= ~(FLEXIO_CTRL_DOZEN_MASK | FLEXIO_CTRL_DBGE_MASK | FLEXIO_CTRL_FASTACC_MASK | FLEXIO_CTRL_FLEXEN_MASK);
    ctrlReg |= (FLEXIO_CTRL_DBGE(masterConfig->enableInDebug) | FLEXIO_CTRL_FASTACC(masterConfig->enableFastAccess) |
                FLEXIO_CTRL_FLEXEN(masterConfig->enableMaster));
    if (!masterConfig->enableInDoze)
    {
        ctrlReg |= FLEXIO_CTRL_DOZEN_MASK;
    }

    base->flexioBase->CTRL = ctrlReg;

    /* Do hardware configuration. */
    /* 1. Configure the shifter 0 for tx. */
    shifterConfig.timerSelect = base->timerIndex[0];
    shifterConfig.pinConfig   = kFLEXIO_PinConfigOutput;
    shifterConfig.pinSelect   = base->SDOPinIndex;
    shifterConfig.pinPolarity = pinPolarity;
    shifterConfig.shifterMode = kFLEXIO_ShifterModeTransmit;
    shifterConfig.inputSource = kFLEXIO_ShifterInputFromPin;
    if (masterConfig->phase == kFLEXIO_SPI_ClockPhaseFirstEdge)
    {
        shifterConfig.timerPolarity = kFLEXIO_ShifterTimerPolarityOnNegitive;
        shifterConfig.shifterStop   = kFLEXIO_ShifterStopBitDisable;
        shifterConfig.shifterStart  = kFLEXIO_ShifterStartBitDisabledLoadDataOnEnable;
    }
    else
    {
        shifterConfig.timerPolarity = kFLEXIO_ShifterTimerPolarityOnPositive;
        shifterConfig.shifterStop   = kFLEXIO_ShifterStopBitLow;
        shifterConfig.shifterStart  = kFLEXIO_ShifterStartBitDisabledLoadDataOnShift;
    }

    FLEXIO_SetShifterConfig(base->flexioBase, base->shifterIndex[0], &shifterConfig);

    /* 2. Configure the shifter 1 for rx. */
    shifterConfig.timerSelect  = base->timerIndex[0];
    shifterConfig.pinConfig    = kFLEXIO_PinConfigOutputDisabled;
    shifterConfig.pinSelect    = base->SDIPinIndex;
    shifterConfig.pinPolarity  = pinPolarity;
    shifterConfig.shifterMode  = kFLEXIO_ShifterModeReceive;
    shifterConfig.inputSource  = kFLEXIO_ShifterInputFromPin;
    shifterConfig.shifterStop  = kFLEXIO_ShifterStopBitDisable;
    shifterConfig.shifterStart = kFLEXIO_ShifterStartBitDisabledLoadDataOnEnable;
    if (masterConfig->phase == kFLEXIO_SPI_ClockPhaseFirstEdge)
    {
        shifterConfig.timerPolarity = kFLEXIO_ShifterTimerPolarityOnPositive;
    }
    else
    {
        shifterConfig.timerPolarity = kFLEXIO_ShifterTimerPolarityOnNegitive;
    }

    FLEXIO_SetShifterConfig(base->flexioBase, base->shifterIndex[1], &shifterConfig);

    /*3. Configure the timer 0 for SCK. */
    timerConfig.triggerSelect   = FLEXIO_TIMER_TRIGGER_SEL_SHIFTnSTAT(base->shifterIndex[0]);
    timerConfig.triggerPolarity = triggerPolarity;
    timerConfig.triggerSource   = kFLEXIO_TimerTriggerSourceInternal;
    timerConfig.pinConfig       = kFLEXIO_PinConfigOutput;
    timerConfig.pinSelect       = base->SCKPinIndex;
    timerConfig.pinPolarity     = pinPolarity;
    timerConfig.timerMode       = kFLEXIO_TimerModeDual8BitBaudBit;
    timerConfig.timerOutput     = kFLEXIO_TimerOutputZeroNotAffectedByReset;
    timerConfig.timerDecrement  = kFLEXIO_TimerDecSrcOnFlexIOClockShiftTimerOutput;
    timerConfig.timerReset      = kFLEXIO_TimerResetNever;
    timerConfig.timerDisable    = kFLEXIO_TimerDisableOnTimerCompare;
    timerConfig.timerEnable     = kFLEXIO_TimerEnableOnTriggerHigh;
    timerConfig.timerStop       = kFLEXIO_TimerStopBitEnableOnTimerDisable;
    timerConfig.timerStart      = kFLEXIO_TimerStartBitEnabled;

    timerDiv = srcClock_Hz / masterConfig->baudRate_Bps;
    timerDiv = timerDiv / 2 - 1;

    timerCmp = ((uint32_t)(masterConfig->dataMode * 2 - 1U)) << 8U;
    timerCmp |= timerDiv;

    timerConfig.timerCompare = timerCmp;

    FLEXIO_SetTimerConfig(base->flexioBase, base->timerIndex[0], &timerConfig);

    /* 4. Configure the timer 1 for CSn. */
    timerConfig.triggerSelect   = FLEXIO_TIMER_TRIGGER_SEL_TIMn(base->timerIndex[0]);
    timerConfig.triggerPolarity = triggerPolarity;
    timerConfig.triggerSource   = kFLEXIO_TimerTriggerSourceInternal;
    timerConfig.pinConfig       = kFLEXIO_PinConfigOutput;
    timerConfig.pinSelect       = base->CSnPinIndex;
    timerConfig.pinPolarity     = pinPolarity;
    timerConfig.timerMode       = kFLEXIO_TimerModeSingle16Bit;
    timerConfig.timerOutput     = kFLEXIO_TimerOutputOneNotAffectedByReset;
    timerConfig.timerDecrement  = kFLEXIO_TimerDecSrcOnFlexIOClockShiftTimerOutput;
    timerConfig.timerReset      = kFLEXIO_TimerResetNever;
    timerConfig.timerDisable    = kFLEXIO_TimerDisableOnPreTimerDisable;
    timerConfig.timerEnable     = kFLEXIO_TimerEnableOnPrevTimerEnable;
    timerConfig.timerStop       = kFLEXIO_TimerStopBitDisabled;
    timerConfig.timerStart      = kFLEXIO_TimerStartBitDisabled;

    timerConfig.timerCompare = 0xFFFFU;

    FLEXIO_SetTimerConfig(base->flexioBase, base->timerIndex[1], &timerConfig);
}


sl_status_t sl_wfx_host_init_bus(void)
{
  /* Init SPI interface */
  
    gpio_pin_config_t wf200_CS_GPO = {
        kGPIO_DigitalOutput,
        0, // Default value.
    };


    gpio_pin_config_t wf200_RESET_GPO = {
        kGPIO_DigitalOutput,
        1,  // Default value is "HIGH", otherwise causes a 1 uSec glitch on the GPO
    };
   
    gpio_pin_config_t wf200_WUP_GPO = {
        kGPIO_DigitalOutput,
        1, // Default value.
    };
    
    gpio_pin_config_t wf200_WIRQ_gpi = {
        kGPIO_DigitalInput,
        0, // This value is ignored for inputs.
    };

    gpio_pin_config_t wf200_spi_IRQ_gpi = {
        kGPIO_DigitalInput,
        0, // This value is ignored for inputs.
    };

    gpio_interrupt_config_t wf200_spi_IRQ_gpi_int_config = { 
         kGPIO_PinIntEnableEdge, 
         kGPIO_PinIntEnableHighOrRise // kGPIO_PinIntEnableLowOrFall // 
    };
    

    flexio_spi_master_config_t masterConfig;
    
    /* Clock setting for Flexio */
    CLOCK_AttachClk(FLEXIO_CLOCK_SRC);
    CLOCK_SetClkDiv(kCLOCK_DivFlexioClk, FLEXIO_CLOCK_DIVIDER);

    RESET_ClearPeripheralReset(kFLEXIO_RST_SHIFT_RSTn);
  
    /* Master config */
    masterConfig.enableMaster = true;
    masterConfig.enableInDoze = false;
    masterConfig.enableInDebug = true;
    masterConfig.enableFastAccess = false;
    masterConfig.baudRate_Bps = 2000000U;
    masterConfig.phase = kFLEXIO_SPI_ClockPhaseFirstEdge;
    masterConfig.dataMode = kFLEXIO_SPI_8BitMode;

    spiDev.flexioBase      = BOARD_FLEXIO_BASE;
    spiDev.SDOPinIndex     = FLEXIO_SPI_MOSI_PIN;
    spiDev.SDIPinIndex     = FLEXIO_SPI_MISO_PIN;
    spiDev.SCKPinIndex     = FLEXIO_SPI_SCK_PIN;
    spiDev.CSnPinIndex     = 1; // FLEXIO_SPI_PCS0_PIN;
    spiDev.shifterIndex[0] = 0U;
    spiDev.shifterIndex[1] = 1U;
    spiDev.timerIndex[0]   = 0U;
    spiDev.timerIndex[1]   = 1U;

    //FLEXIO_SPI_EnableDMA(&spiDev, kFLEXIO_SPI_TxDmaEnable | kFLEXIO_SPI_RxDmaEnable,  true );

    //NVIC_SetPriority(MASTER_SPI_IRQ, 1U);

    EnableIRQ(GPIO_INTA_IRQn);
    GPIO_PortInit_noreset(GPIO, SL_WFX_RESET_PORT);
    GPIO_PinInit(GPIO, SL_WFX_IRQ_PORT_SPI, SL_WFX_IRQ_GPIO_SPI, &wf200_spi_IRQ_gpi);
    GPIO_SetPinInterruptConfig(GPIO, SL_WFX_IRQ_PORT_SPI, SL_WFX_IRQ_GPIO_SPI, &wf200_spi_IRQ_gpi_int_config);
    
    GPIO_PinInit(GPIO, SL_WFX_RESET_PORT, SL_WFX_RESET_GPIO, &wf200_RESET_GPO);

    GPIO_PortInit_noreset(GPIO, SL_WFX_WUP_PORT);
    GPIO_PinInit(GPIO, SL_WFX_WUP_PORT, SL_WFX_WUP_GPIO, &wf200_WUP_GPO);
 
    GPIO_PortInit_noreset(GPIO, SL_WFX_WIRQ_PORT);
    GPIO_PinInit(GPIO, SL_WFX_WIRQ_PORT, SL_WFX_WIRQ_GPIO, &wf200_WIRQ_gpi);

    // CS is controlled manually.
    GPIO_PortInit_noreset(GPIO, SL_WFX_CS_PORT_SPI);
    GPIO_PinInit(GPIO, SL_WFX_CS_PORT_SPI, SL_WFX_CS_GPIO_SPI, &wf200_CS_GPO);

    for ( volatile uint32_t i ; i < 0xffffff ; i++ )
    {
       __asm("NOP"); /* delay */
    }
        

    FLEXIO_SPI_MasterInit_for_wfx(&spiDev, &masterConfig, FLEXIO_CLOCK_FREQUENCY);
    
    // /* Set up master transfer */
    FLEXIO_SPI_MasterTransferCreateHandle(&spiDev, &g_m_handle, FLEXIO_SPI_MasterUserCallback, NULL);

  /*Create semaphore to handle SPI*/
//  spiDMASemaphore = xSemaphoreCreateBinary();
//  xSemaphoreGive(spiDMASemaphore);  
  return SL_STATUS_OK;
}

sl_status_t sl_wfx_host_deinit_bus(void)
{
  /* Deinit SPI interface */
  //EA:ToDo
  return SL_STATUS_OK;
}

sl_status_t sl_wfx_host_spi_cs_assert(void)
{
  GPIO_PinWrite(GPIO, SL_WFX_CS_PORT_SPI, SL_WFX_CS_GPIO_SPI, 0);

  return SL_STATUS_OK;
}

sl_status_t sl_wfx_host_spi_cs_deassert(void)
{
  GPIO_PinWrite(GPIO, SL_WFX_CS_PORT_SPI, SL_WFX_CS_GPIO_SPI, 1); 
  
  return SL_STATUS_OK;
}

sl_status_t sl_wfx_host_spi_transfer_no_cs_assert(sl_wfx_host_bus_transfer_type_t type,
                                                  uint8_t *header,
                                                  uint16_t header_length,
                                                  uint8_t *buffer,
                                                  uint16_t buffer_length)
{

#ifdef HOST_SPI_DEBUG
	  printf("[DEBUG] %s\n", __func__);
#endif

  sl_status_t    result  = SL_STATUS_FAIL;
  const bool     is_read = ( type == SL_WFX_BUS_READ );

  uint8_t flags = kFLEXIO_SPI_8bitMsb;

  flexio_spi_transfer_t headermasterBuffer;
  flexio_spi_transfer_t rxmasterBuffer;
  flexio_spi_transfer_t txmasterBuffer;

  /* Start master transfer */
  headermasterBuffer.txData = header;
  headermasterBuffer.rxData = NULL;
  headermasterBuffer.dataSize = header_length;
  headermasterBuffer.flags = flags;

  FLEXIO_SPI_MasterTransferBlocking(&spiDev, &headermasterBuffer);
 
	if (is_read)
	{

		rxmasterBuffer.txData = NULL;
		rxmasterBuffer.rxData = buffer;
		rxmasterBuffer.dataSize = buffer_length;
		rxmasterBuffer.flags = flags;
		FLEXIO_SPI_MasterTransferBlocking(&spiDev, &rxmasterBuffer);
	    result = SL_STATUS_OK;
        
	}else
	{
		txmasterBuffer.txData = buffer;
		txmasterBuffer.rxData = NULL;
		txmasterBuffer.dataSize = buffer_length;
		txmasterBuffer.flags = flags;
		FLEXIO_SPI_MasterTransferBlocking(&spiDev, &txmasterBuffer);
	    result = SL_STATUS_OK;

	}
  return result;

}

sl_status_t sl_wfx_host_enable_platform_interrupt(void)
{
   GPIO_PinEnableInterrupt(GPIO, SL_WFX_IRQ_PORT_SPI, 
                                 SL_WFX_IRQ_GPIO_SPI, 
                                 SPI_WIRQ_INTERRUPT_ID);
  return SL_STATUS_OK;

}

sl_status_t sl_wfx_host_disable_platform_interrupt(void)
{

  GPIO_PinDisableInterrupt(GPIO, SL_WFX_IRQ_PORT_SPI, SL_WFX_IRQ_GPIO_SPI, SPI_WIRQ_INTERRUPT_ID);

  return SL_STATUS_OK;
}

