/*
 * Java
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * Use of this source code is governed by a BSD-style license that can be found with this software.
 */
package com.microej.demo.watch.util.path;

import static com.verisilicon.vivante.vg_lite_gradient.vg_lite_init_grad;
import static com.verisilicon.vivante.vg_lite_gradient.vg_lite_set_grad;
import static com.verisilicon.vivante.vg_lite_gradient.vg_lite_update_grad;

import com.verisilicon.vivante.vg_lite_gradient;

/**
 * Represents the gradient information: colors, colors anchor points, and angle.
 */
public class GradientStyle {

	/** The default gradient style to use when none was specified. */
	public static final GradientStyle DEFAULT_GRADIENT_STYLE = new GradientStyle(new int[] { 0xFFFFFFFF, 0x00FFFFFF },
			new int[] { 0, 255 }, 0);

	private final int[] colors;

	private final int[] stops;

	private final int angle;

	/**
	 * Creates a gradient style, given the colors to use, the position of colors in the gradient and the gradient angle.
	 *
	 * <p>
	 * The positions of the colors are relative to the gradient length from 0 (start) to 255 (end).
	 *
	 * @param colors
	 *            the colors to use.
	 * @param stops
	 *            the positions of colors in the gradient.
	 * @param angle
	 *            the gradient angle.
	 */
	public GradientStyle(int[] colors, int[] stops, int angle) {
		this.colors = colors.clone();
		this.stops = stops.clone();
		this.angle = angle;
	}

	/**
	 * Gets the number of colors in this gradient.
	 *
	 * @return the number of colors of this gradient.
	 */
	public int getColorCount() {
		return this.colors.length;
	}

	/**
	 * Gets the colors.
	 *
	 * <p>
	 * The returned array is a shallow copy of the array field.
	 *
	 * @return the colors.
	 */
	public int[] getColors() {
		return this.colors.clone();
	}

	/**
	 * Gets the stops.
	 *
	 * <p>
	 * The returned array is a shallow copy of the array field.
	 *
	 * @return the stops.
	 */
	public int[] getStops() {
		return this.stops.clone();
	}

	/**
	 * Gets the angle.
	 *
	 * @return the angle.
	 */
	public int getAngle() {
		return this.angle;
	}

	/**
	 * Apply style on the given VGLite gradient.
	 *
	 * @param gradient
	 *            the gradient where applying the style.
	 */
	public void applyStyle(vg_lite_gradient gradient) {
		vg_lite_init_grad(gradient);
		vg_lite_set_grad(gradient, getColorCount(), getColors(), getStops());
		vg_lite_update_grad(gradient);
	}

}
