/*
 * Java
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * Use of this source code is governed by a BSD-style license that can be found with this software.
 */
package com.microej.demo.watch.util.path;

import java.io.Closeable;

import com.microej.vivante.utils.SvgImage;
import com.verisilicon.vivante.vg_lite_matrix;

import ej.annotation.NonNull;
import ej.annotation.Nullable;
import ej.bon.IllegalStateException;
import ej.microui.display.GraphicsContext;

/**
 * A vector graphics holds a {@link vg_lite_matrix} and a {@link SvgImage}. The image is composed of one or more paths.
 *
 * <p>
 * The vector graphics render the image using the matrix.
 */
public abstract class VectorGraphics implements Closeable {

	private final SvgImage image;

	private vg_lite_matrix pathMatrix;

	/**
	 * Creates a new vector graphics.
	 *
	 * <p>
	 * This method creates a new matrix that client code can use to apply transformations on this vector graphics.
	 *
	 * <p>
	 * Calling this constructor results in a dynamic memory allocation in low-level VGLite APIs to store the matrix
	 * information. One must call method {@link #close()} before dereferencing this object in order to free memory.
	 *
	 * @param image
	 *            the SVG vector image that compose this vector graphics.
	 */
	public VectorGraphics(@NonNull SvgImage image) {
		this.pathMatrix = new vg_lite_matrix();
		this.image = image;
	}

	/**
	 * Updates the path matrices.
	 *
	 * <p>
	 * Call this method when the transformation parameters has changed (i.e scale factor, rotation angle) and you want
	 * the vector graphics to reflect that change.
	 *
	 * <p>
	 * This method should be called in the UI thread to avoid concurrency issues.
	 */
	protected void updateMatrices() {
		// does nothing by default, subclasses may override this behavior to transform the vector graphics.
	}

	/**
	 * Renders the vector graphics on the given {@link GraphicsContext}.
	 *
	 * <p>
	 * The default implementation renders the SVG image applying the matrix on it.
	 *
	 * @param g
	 *            the graphics context where to render the content of the widget.
	 */
	public void render(GraphicsContext g) {
		getImage().draw(g, getMatrix());
	}

	/**
	 * Gets the SVG image.
	 *
	 * @return the image.
	 */
	@NonNull
	protected SvgImage getImage() {
		return this.image;
	}

	/**
	 * Gets the matrix.
	 *
	 * <p>
	 * Use the matrix returned by this method to apply transformations to your vector graphics.
	 *
	 * @return the matrix.
	 * @throws IllegalStateException
	 *             if this vector graphics has been closed (see {@link #close()}.
	 */
	@Nullable
	protected vg_lite_matrix getMatrix() throws IllegalStateException {
		vg_lite_matrix matrix = this.pathMatrix;
		if (matrix == null) {
			throw new IllegalStateException();
		}

		return matrix;
	}

	/**
	 * Closes this vector graphics and frees any allocated memory used for matrix.
	 *
	 * <p>
	 * Once the vector graphics has been closed, further {@link #getMatrix()} invocations will throw an
	 * {@link IllegalStateException}.
	 */
	@Override
	public void close() {
		vg_lite_matrix pathMatrix = this.pathMatrix;
		if (pathMatrix != null) {
			pathMatrix.close();
			this.pathMatrix = null;
		}
	}
}
