/*
 * Java
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * Use of this source code is governed by a BSD-style license that can be found with this software.
 */
package com.microej.path.chart;

import com.microej.demo.watch.util.path.GradientPathGraphics;
import com.microej.demo.watch.util.path.GradientStyle;
import com.microej.vivante.utils.GradientPath;
import com.microej.vivante.utils.Path;
import com.microej.vivante.utils.PathBufferFP32;
import com.verisilicon.vivante.vg_lite;
import com.verisilicon.vivante.vg_lite_gradient;
import com.verisilicon.vivante.vg_lite_matrix;
import com.verisilicon.vivante.vg_lite_path;

/**
 * A vector graphics that represents the gradient under the curve of a line chart.
 *
 * <p>
 * The path is built using Bézier curve from the data array in input.
 *
 */
public class ChartGradient extends GradientPathGraphics {

	private static final float GRADIENT_SIZE = 256f;

	private static final int SMOOTH_FACTOR = 3;

	private static final int TOP_PADDING = 5;

	private final int width;

	private final int height;

	private final int hStep;

	/**
	 * Constructs the vector curve, given the data array to plot.
	 *
	 * @param y
	 *            the top y-coordinate of the vector graphics' bounding-box.
	 * @param height
	 *            the height of the the vector graphics' bounding-box.
	 * @param data
	 *            the data array which contains the values to plot.
	 * @param step
	 *            the horizontal spacing to use between each data point.
	 * @param gradientStyle
	 *            the gradient information.
	 */
	public ChartGradient(int y, int height, float[] data, int step, GradientStyle gradientStyle) {
		super(createPath(data, step, height), gradientStyle);
		this.hStep = step;
		this.width = step * data.length;
		this.height = height;
		initialize(y, height, gradientStyle.getAngle(), y + height - TOP_PADDING);
	}

	private static GradientPath createPath(float[] data, int hStep, int height) {

		PathBufferFP32 pathBuffer = new PathBufferFP32();
		int width = computePath(pathBuffer, data, hStep);
		vg_lite_path path = pathBuffer.toVGLitePath(0, -height, width, 0, vg_lite_path.VG_LITE_HIGH);

		return new GradientPath(path, new vg_lite_gradient(), vg_lite.VG_LITE_FILL_EVEN_ODD);
	}

	private static int computePath(PathBufferFP32 pathBuffer, float[] data, int hStep) {
		pathBuffer.move(0, 0);

		int i = 0;
		int x = hStep;
		int width = hStep * data.length;

		float firstData = getValueAtIndex(data, i);
		double sectionAngle = getSectionAngle(getValueAtIndex(data, i - 1), firstData, getValueAtIndex(data, i + 1),
				hStep);
		i++;

		pathBuffer.line(0, -firstData);

		while (true) {
			float prev = getValueAtIndex(data, i - 1);
			float cur = getValueAtIndex(data, i);
			float next = getValueAtIndex(data, i + 1);

			sectionAngle = plot(pathBuffer, sectionAngle, prev, cur, next, x, hStep);

			if (x >= (width)) {
				break;
			}

			i++;
			x += hStep;
		}

		pathBuffer.line(x, 0);
		pathBuffer.close();
		return width;
	}

	private static double plot(PathBufferFP32 pb, double prevAngle, float prev, float cur, float next, int x,
			int hStep) {
		double tmpPrevAngle = Math.atan2(prev - cur, hStep);
		double tmpNextAngle = Math.atan2(next - cur, hStep);
		double sectionAngle = (-tmpPrevAngle + tmpNextAngle) / 2;

		float cx1 = (float) (x - hStep + hStep * Math.cos(prevAngle) / SMOOTH_FACTOR);
		float cy1 = (float) (prev + hStep * Math.sin(prevAngle) / SMOOTH_FACTOR);
		float cx2 = (float) (x - hStep * Math.cos(sectionAngle) / SMOOTH_FACTOR);
		float cy2 = (float) (cur - hStep * Math.sin(sectionAngle) / SMOOTH_FACTOR);
		pb.cubic(cx1, -cy1, cx2, -cy2, x, -cur);

		return sectionAngle;
	}

	private static double getSectionAngle(float prev, float cur, float next, int hStep) {
		double tmpPrevAngle = Math.atan2(prev - cur, hStep);
		double tmpNextAngle = Math.atan2(next - cur, hStep);

		return (-tmpPrevAngle + tmpNextAngle) / 2;
	}

	private void initialize(int y, int height, float gradientAngle, int translateY) {
		vg_lite_matrix matrix = getMatrix();
		vg_lite_matrix.vg_lite_identity(matrix);
		vg_lite_matrix.vg_lite_translate(0, translateY, matrix);

		vg_lite_gradient gradient = getGradient();
		vg_lite_matrix gradientMatrix = gradient.getMatrix();
		vg_lite_matrix.vg_lite_identity(gradientMatrix);
		vg_lite_matrix.vg_lite_translate(0, y, gradientMatrix);
		vg_lite_matrix.vg_lite_rotate(gradientAngle, gradientMatrix);
		vg_lite_matrix.vg_lite_scale(height / GRADIENT_SIZE, 1f, gradientMatrix);
	}

	private static float getValueAtIndex(float[] data, int index) {
		int arrayLength = data.length;
		while (index < 0) {
			index += arrayLength;
		}

		while (index >= arrayLength) {
			index -= arrayLength;
		}

		return data[index];
	}

	/**
	 * Sets the data to plot.
	 *
	 * <p>
	 * This method updates the vector graphics' path to represent the specified data points.
	 *
	 * <p>
	 * This method should be called in the UI thread.
	 *
	 * @param data
	 *            the data to plot.
	 *
	 */
	public void setData(float[] data) {
		PathBufferFP32 pathBuffer = new PathBufferFP32();

		computePath(pathBuffer, data, this.hStep);

		Path path = getPath();
		assert path != null;
		pathBuffer.updateVGLitePath(path.getVGLitePath(), 0, -this.height, this.width, 0, vg_lite_path.VG_LITE_HIGH);
	}

}
