/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.annotation.processor.ant;

import java.io.File;
import java.util.logging.SimpleFormatter;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.taskdefs.Javac;
import org.apache.tools.ant.taskdefs.Javac.ImplementationSpecificArgument;
import org.apache.tools.ant.types.Path;
import org.apache.tools.ant.types.Reference;

/**
 *
 */
public class GIMLAnnotationProcessorTask extends Task {

	private static final String ANT_TASK_NAME = "codegeneration"; // to synchronize with antlib.xml file
	private static final String GENERATED_SOURCES_OUTPUT_DIR_OPTION = "-s";
	private static final String SIMPLEFORMATTER_CLASS_NAME = SimpleFormatter.class.getCanonicalName();
	private static final String LOG_FORMATTER_FORMAT = "%4$-6s %5$s%6$s%n";

	public static final boolean DEBUG_MODE = Boolean.getBoolean("GIMLAnnotationProcessor.debug");

	private Javac _javacTask; // NOSONAR
	private File _destDir; // NOSONAR
	private Reference _sourcePathReference;
	private Reference _classpathReference;
	private String _encoding;

	@Override
	public String getTaskName() {
		return ANT_TASK_NAME;
	}

	@Override
	public void execute() throws BuildException {
		try {
			super.execute();

			// Check all is green and ready to go!
			if (this._destDir == null) {
				throw new BuildException("'destdir' attribute must be set.");
			}
			if (this._sourcePathReference == null) {
				throw new BuildException("'sourcepathref' attribute must be set.");
			}

			if (this._classpathReference == null) {
				throw new BuildException("'classpathref' attribute must be set.");
			}

			this._javacTask = new Javac();
			this._javacTask.setProject(getProject());

			// Create compiler options
			this.createCompilerArg(
					GENERATED_SOURCES_OUTPUT_DIR_OPTION + " \"" + this._destDir.getAbsolutePath() + "\"");

			if (!DEBUG_MODE) {
				// Default mode is forked for 2 reasons:
				// - to execute a System.exit to end generation
				// - to set the following formatting option
				this.createCompilerArg(
						"-J-D" + SIMPLEFORMATTER_CLASS_NAME + ".format=\"" + LOG_FORMATTER_FORMAT + "\"");
				this._javacTask.setFork(true);
			}

			// Set other Javac options
			if (this._encoding != null) {
				this._javacTask.setEncoding(this._encoding);
			}
			this._javacTask.setIncludeantruntime(false);
			Path sourcePath = new Path(getProject());
			sourcePath.setRefid(this._sourcePathReference);
			Path classpath = new Path(getProject());
			classpath.setRefid(this._classpathReference);

			this._javacTask.setSrcdir(sourcePath);
			this._javacTask.setClasspath(classpath);
			this._javacTask.setCreateMissingPackageInfoClass(false);
			this._javacTask.setTaskName(this.getTaskName()); // for logger to have the task name
			this._javacTask.setSource("1.7");

			// Finally, invoke javac
			this._javacTask.execute();
		} catch (Throwable e) {
			throw new BuildException(e);
		}
	}

	/**
	 * Adds a reference to a classpath defined elsewhere.
	 *
	 * @param r
	 *            a reference to a classpath
	 */
	public void setClasspathRef(final Reference r) {
		this._classpathReference = r;
	}

	/**
	 * Set the destination directory into which the Java source files should be generated.
	 *
	 * @param destDir
	 *            the destination directory
	 */
	public void setDestDir(final File destDir) {
		this._destDir = destDir;
	}

	/**
	 * Set the Java source file encoding name.
	 *
	 * @param encoding
	 *            the source file encoding
	 */
	public void setEncoding(final String encoding) {
		this._encoding = encoding;
	}

	/**
	 * Adds a reference to a source path defined elsewhere.
	 *
	 * @param r
	 *            a reference to a source path
	 */
	public void setSourcepathRef(final Reference r) {
		this._sourcePathReference = r;
	}

	private void createCompilerArg(String value) {
		ImplementationSpecificArgument arg = this._javacTask.createCompilerArg();
		arg.setLine(value);
	}
}
