/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.binary;

import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;


/**
 * Map from a GIML object identifier to the GIML object address.<p>
 * An object identifier may be a String.<p>
 * Usage of this class consists in:
 * <ol>
 * 	<li>creating a new instance</li>
 * 	<li>adding the objects with the {@link #put(Object, int, int)} method</li>
 * 	<li>completing the map with the {@link #complete()} method</li>
 * 	<li>write the result in an {@link OutputStream} with the {@link #writeTo(OutputStream, int)} method</li>
 * </ol>
 */
public class GIMLObjectMap {

	private static final boolean LOG = false;

	/**
	 * Registry that gives for an identifier the corresponding GIML object reference. The identifier can be
	 * any kind of object.
	 */
	private final Map<Object, GIMLObjectReference> objectMap;

	/** @see #GIMLObjectMap(float) */
	private final float loaderFactor;

	/** Set to true after a call to {@link #complete()} */
	private boolean completed;

	/** Entries of the map computed during {@link #complete()}. */
	private Bucket[] buckets;

	/**
	 * Constructs an empty GIMLObjectMap with the specified load factor.<p>
	 * The load factor is a measure of how full the hash table is allowed to get before its capacity is automatically increased.
	 * Higher values decrease the space overhead but increase the lookup cost.
	 * 
	 * @param loadFactor the load factor
	 */
	public GIMLObjectMap(float loadFactor) {
		this.objectMap = new HashMap<>();
		this.loaderFactor = loadFactor;
		this.completed = false;
	}

	/**
	 * Add the given object in the map.
	 * @param identifier the identifier of the object.
	 * @param identifierAddress address of the identifier in the constant pool (relative from the constant pool start).
	 * @param objectAddress address of the object (relative from the data start).
	 * 
	 * @throws IllegalArgumentException if there is already an entry for the given identifierAddress.
	 * @throws IllegalStateException if this map has been completed.
	 */
	public void put(Object identifier, int identifierAddress, int objectAddress){
		checkNotCompleted();
		if(objectMap.containsKey(identifier)){
			throw new IllegalArgumentException("Duplicated entry: "+identifier);
		}

		objectMap.put(identifier, new GIMLObjectReference(identifierAddress, objectAddress));
	}

	/**
	 * Completes this map and compute the size in bytes of the serialized version of the map.<p>
	 * After this call, any call to {@link #put(Object, int, int)} will throws an {@link IllegalStateException}.
	 * @return the size of this map in bytes once it is serialized.
	 */
	public int complete(){
		checkNotCompleted();

		float loaderFactor = this.loaderFactor;

		while(!completed){
			try{
				Bucket[] buckets = createAndFillBuckets(loaderFactor);
				this.buckets = buckets;
				this.completed = true;
			}
			catch(IOException e){
				//At least one bucket is full with this loaderFactor.
				//Try with a smaller loaderFactor
				loaderFactor = loaderFactor/2;
				if(loaderFactor < 0.25f){
					throw new InternalError("GIML object table size limit reached. Too many entries with the same hashcode.");
				}
			}
		}

		if(LOG){
			System.out.println(getStatus());
		}

		return computeSerializedTableSize(buckets);
	}

	private Bucket[] createAndFillBuckets(float loaderFactor) throws IOException{
		// Compute the number of buckets
		int mapSize = objectMap.size();
		int bucketsTableLength; // number of buckets

		if(mapSize == 0){
			bucketsTableLength = 0;
		}
		else {
			bucketsTableLength = (int)(mapSize / loaderFactor);
			if(bucketsTableLength == 0){
				bucketsTableLength = 1;
			}
		}

		// Creates the buckets
		Bucket[] buckets = new Bucket[bucketsTableLength];
		for(int i=bucketsTableLength ; --i>=0 ;){
			buckets[i] = new Bucket();
		}

		// Fill the buckets
		for(Entry<Object, GIMLObjectReference> entry : objectMap.entrySet()){
			int hash = entry.getKey().hashCode();
			int index = Math.abs(hash % bucketsTableLength);
			// May throw an IOException if the bucket is full
			buckets[index].addEntry(entry.getValue());
		}

		return buckets;
	}

	/**
	 * Computes the size of an object map in bytes with the given buckets.
	 * @param buckets the buckets of the map.
	 * @return a size in bytes.
	 */
	private int computeSerializedTableSize(Bucket[] buckets){
		final int sizeOfInt24 = 3;// size of an 24-bits integer
		final int sizeOfInt16 = 2;// size of an 16-bits integer
		final int sizeOfInt8 = 1;// size of an 8-bits integer

		final int headerSize = sizeOfInt16; // table length encoded on 16-bits
		final int bucketsTableSize = buckets.length * sizeOfInt16;

		int size = headerSize + bucketsTableSize;

		//WARNING: keep this order
		for(int i=0 ; i<buckets.length ; i++){
			Bucket bucket = buckets[i];
			//set the address of the bucket now it is needed to generate the table
			bucket.setAddress(size);

			final int bucketHeaderSize = sizeOfInt8; // bucket length encoded on 8-bits
			final int bucketEntrySize = sizeOfInt24 + sizeOfInt24; // 24-bits for the stringOffset + 24-bits for the objectOffset

			int bucketSize =  bucketHeaderSize + bucketEntrySize*bucket.size();

			size += bucketSize;
		}

		return size;
	}

	/**
	 * Checks that this map is not completed. If the map is completed (i.e. a call to {@link #complete()}
	 * has been done) then an {@link IllegalStateException} is thrown.
	 */
	private void checkNotCompleted(){
		if(completed){
			throw new IllegalStateException("Operation not allowed: object Map has been completed.");
		}
	}

	/**
	 * Serialize the content of this map in the given {@link OutputStream}.
	 * @param out the output stream to which to write the data.
	 * @exception IOException if an I/O error occurs.
	 */
	public void writeTo(OutputStream out) throws IOException{
		DataOutputStream dataOut = new DataOutputStream(out);

		// Header
		int tableLength = buckets.length;
		checkSizeLimit(tableLength, Bits._16BITS);
		dataOut.writeChar(tableLength);

		// Bucket pointer table
		//WARNING: keep this order
		for(int i=0 ; i<buckets.length ; i++){
			Bucket bucket = buckets[i];

			int bucketAddress = bucket.getAddress();
			checkSizeLimit(tableLength, Bits._16BITS);
			dataOut.writeChar(bucketAddress);
		}

		// Bucket content table
		//WARNING: keep this order
		for(int i=0 ; i<buckets.length ; i++){
			Bucket bucket = buckets[i];

			int nbEntries = bucket.size();
			checkSizeLimit(nbEntries, Bits._8BITS);
			dataOut.writeByte(nbEntries);

			//WARNING: keep this order
			for(int entryIndex=0 ; entryIndex<nbEntries ; entryIndex++){
				GIMLObjectReference gimlObject = bucket.getEntry(entryIndex);
				int stringOffset = gimlObject.identifierAddress;
				write24BitsValue(dataOut, stringOffset);
				int objectOffset = gimlObject.objectAddress;
				write24BitsValue(dataOut, objectOffset);
			}

		}

	}

	/**
	 * Writes the given value on 24-bits in the given stream in big-endian.<p>
	 * Tests if the value can be stored on 24-bits and throws an
	 * {@link InternalError} if it cannot.
	 * @param out the stream in which to write the value
	 * @param value the value to write
	 */
	private void write24BitsValue(OutputStream out, int value) throws IOException{
		checkSizeLimit(value, Bits._24BITS);
		out.write(value >> 16);
		out.write(value >> 8);
		out.write(value);

	}


	/**
	 * Call {@link #checkSizeLimit(int, Bits, String)} with a default message.
	 */
	private void checkSizeLimit(int valueToTest, Bits nbBits){
		checkSizeLimit(valueToTest, nbBits, "GIML object table size limit reached.");
	}

	/**
	 * Checks if the given value can be stored on nbBits bits.
	 * If the check fails, then an internal error with the given message is thrown.
	 * @param valueToTest the value to test
	 * @param nbBits number of bits in which we want to store the value.
	 * @param errorMessage the message of the exception thrown if the check fails
	 * 
	 * @throws InternalError if the check fails
	 */
	private void checkSizeLimit(int valueToTest, Bits nbBits, String errorMessage){
		int mask = 0;
		switch(nbBits){
			case _8BITS:{
				mask = 0xFF;
				break;
			}
			case _16BITS:{
				mask = 0xFFFF;
				break;
			}
			case _24BITS:{
				mask = 0xFFFFFF;
			}
		}
		int croppedValue = valueToTest & mask;
		if(croppedValue != valueToTest){
			throw new InternalError(errorMessage);
		}
	}

	private String getStatus(){
		StringBuilder sb = new StringBuilder();
		sb.append(String.format("Nb entries: %d\n", objectMap.size()));
		if(buckets != null){
			sb.append(String.format("Nb buckets: %d\n", buckets.length));
			for(int i=0 ; i<buckets.length ; i++){
				sb.append(String.format(" bucket[%d]: %d elements\n", i, buckets[i].size()));
			}
		}
		return sb.toString();
	}

	public static class GIMLObjectReference {

		/** Address of the identifier. */
		public final int identifierAddress;
		/** Address of the object (relative from the data start). */
		public final int objectAddress;

		/**
		 * Creates a {@link GIMLObjectReference} with the given identifierAddress and objectAddress.
		 * @see #identifierAddress
		 * @see #objectAddress
		 */
		public GIMLObjectReference(int identifierAddress, int objectAddress) {
			super();
			this.identifierAddress = identifierAddress;
			this.objectAddress = objectAddress;
		}


	}

	/**
	 * A list of entries with the same index in the map table.
	 * The index is computed from the hashcode of the entry.
	 */
	public static class Bucket {

		/**
		 * By default the maximum size of a bucket is limited to the max value of an
		 * unsigned 8-bits (2^8 - 1).
		 */
		private static final int DEFAULT_MAXIMUM_SIZE = 255;

		/** Maximum number of elements in this bucket. */
		private final int maximumSize;

		/** The entries of this bucket. */
		private final List<GIMLObjectReference> entries;

		/** Address of this bucket. */
		private int address;

		public Bucket(){
			this(DEFAULT_MAXIMUM_SIZE);
		}

		/**
		 * Creates a bucket with the given maximum size. When the maximum size
		 * is reached a call to {@link #addEntry(GIMLObjectReference)} will produce an {@link IOException}.
		 * @param maximumSize maximum size of this bucket.
		 */
		public Bucket(int maximumSize) {
			this.maximumSize = maximumSize;
			this.entries = new ArrayList<>();
		}

		/**
		 * Add a GIML object in this bucket
		 * @param object the object to add
		 * 
		 * @throws IOException if the maximum number of object is exceeded.
		 */
		public void addEntry(GIMLObjectReference object) throws IOException {
			if(entries.size() >= maximumSize){
				throw new IOException("Maximum number of entries reached");
			}
			entries.add(object);
		}

		public GIMLObjectReference getEntry(int offset){
			return entries.get(offset);
		}

		/**
		 * Returns the number of elements in this bucket.
		 */
		public int size(){
			return entries.size();
		}

		/**
		 * Set the address of the bucket. This address is relative from the beginning of
		 * the table.
		 */
		public void setAddress(int address){
			this.address = address;
		}

		/**
		 * Returns the address of the bucket or -1 if not defined.
		 */
		public int getAddress(){
			return address;
		}

	}

	static enum Bits {
		_8BITS, _16BITS, _24BITS
	}
}
