/*
 * Java
 *
 * Copyright 2016-2017 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.binary;

import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Map;

import com.is2t.nls.StringEncoder;

/**
 * Implementations of the {@link GIMLWriter} that generates a GIML binary data in a compact format.
 */
public class GIMLWriterCompacted implements GIMLWriter{

	private static final int MASK_6BITS = (1<<6)-1;
	private static final int MASK_7BITS = (1<<7)-1;
	private static final int MASK_13BITS = (1<<13)-1;
	private static final int MASK_14BITS = (1<<14)-1;
	private static final int MASK_20BITS = (1<<20)-1;
	private static final int MASK_21BITS = (1<<21)-1;
	private static final int MASK_27BITS = (1<<27)-1;
	private static final int MASK_28BITS = (1<<28)-1;
	private static final long MASK_34BITS = (1l<<34)-1;
	private static final long MASK_41BITS = (1l<<41)-1;
	private static final long MASK_48BITS = (1l<<48)-1;
	private static final long MASK_55BITS = (1l<<56)-1;
	private static final long MASK_62BITS = (1l<<62)-1;

	private static final int TRUE = 1;
	private static final int FALSE = 0;

	/** Size in bytes of the GIML Binary header. */
	private static final int HEADER_SIZE = 12;

	private static final float DEFAULT_OBJECT_MAP_LOAD_FACTOR = 2.0f;

	private final OutputStream out;

	private final WatchableByteArrayOutputStream dataBuffer;
	private final WatchableByteArrayOutputStream constantPoolBuffer;

	private final Map<String, Integer> constantPoolEntries;
	/** String encoder used for non NLS strings. */
	private final StringEncoder stringEncoder;

	private final GIMLObjectMap objectMap;

	public GIMLWriterCompacted(OutputStream out){
		this(out, DEFAULT_OBJECT_MAP_LOAD_FACTOR);
	}

	/**
	 *
	 * @param out the {@link OutputStream} in which the GIML binary is written.
	 * @param objectMapLoadFactor the load factor for the object map (see {@link GIMLObjectMap#GIMLObjectMap(float)})
	 */
	public GIMLWriterCompacted(OutputStream out, float objectMapLoadFactor){
		this.out = out;
		this.dataBuffer = new WatchableByteArrayOutputStream();
		this.constantPoolBuffer = new WatchableByteArrayOutputStream();
		this.constantPoolEntries = new HashMap<>();
		this.stringEncoder = new StringEncoder();
		this.objectMap = new GIMLObjectMap(objectMapLoadFactor);
	}

	@Override
	public void writeVarUInt(int value){
		if((value&MASK_7BITS) != value){
			if((value&MASK_14BITS) != value){
				if((value&MASK_21BITS) != value){
					if((value&MASK_28BITS) != value){
						dataBuffer.write((value>>>27) | 1);
					}
					dataBuffer.write((value>>>20) | 1);
				}
				dataBuffer.write((value>>>13) | 1);
			}
			dataBuffer.write((value>>>6) | 1);
		}
		dataBuffer.write((value<<1) | 0);
	}

	@Override
	public void writeVarSInt(int value){
		if(value == Integer.MIN_VALUE){
			//Special case for MIN_VALUE because -MIN_VALUE is not positive
			dataBuffer.write((Integer.MIN_VALUE >> 27) | 1);
			dataBuffer.write((Integer.MIN_VALUE >> 20) | 1);
			dataBuffer.write((Integer.MIN_VALUE >> 13) | 1);
			dataBuffer.write((Integer.MIN_VALUE >> 6) | 1);
			dataBuffer.write((Integer.MIN_VALUE << 1) | 0);
		}
		else {
			int positiveValue;
			if(value < 0){
				positiveValue = -value-1;
			}
			else {
				positiveValue = value;
			}

			if((positiveValue&MASK_6BITS) != positiveValue){
				if((positiveValue&MASK_13BITS) != positiveValue){
					if((positiveValue&MASK_20BITS) != positiveValue){
						if((positiveValue&MASK_27BITS) != positiveValue){
							dataBuffer.write((value>>27) | 1);
						}
						dataBuffer.write((value>>20) | 1);
					}
					dataBuffer.write((value>>13) | 1);
				}
				dataBuffer.write((value>>6) | 1);
			}
			dataBuffer.write((value<<1) | 0);
		}
	}


	@Override
	public void writeVarLong(long value) {
		if(value == Long.MIN_VALUE){
			//Special case for MIN_VALUE because -MIN_VALUE is not positive
			dataBuffer.write((int) ((Long.MIN_VALUE >> 62) | 1));
			dataBuffer.write((int) ((Long.MIN_VALUE >> 55) | 1));
			dataBuffer.write((int) ((Long.MIN_VALUE >> 48) | 1));
			dataBuffer.write((int) ((Long.MIN_VALUE >> 41) | 1));
			dataBuffer.write((int) ((Long.MIN_VALUE >> 34) | 1));
			dataBuffer.write((int) ((Long.MIN_VALUE >> 27) | 1));
			dataBuffer.write((int) ((Long.MIN_VALUE >> 20) | 1));
			dataBuffer.write((int) ((Long.MIN_VALUE >> 13) | 1));
			dataBuffer.write((int) ((Long.MIN_VALUE >> 6) | 1));
			dataBuffer.write((int) ((Long.MIN_VALUE << 1) | 0));
		}
		else {
			long positiveValue;
			if(value < 0){
				positiveValue = -value;
			}
			else {
				positiveValue = value;
			}

			if((positiveValue&MASK_6BITS) != positiveValue){
				if((positiveValue&MASK_13BITS) != positiveValue){
					if((positiveValue&MASK_20BITS) != positiveValue){
						if((positiveValue&MASK_27BITS) != positiveValue){
							if((positiveValue&MASK_34BITS) != positiveValue){
								if((positiveValue&MASK_41BITS) != positiveValue){
									if((positiveValue&MASK_48BITS) != positiveValue){
										if((positiveValue&MASK_55BITS) != positiveValue){
											if((positiveValue&MASK_62BITS) != positiveValue){
												dataBuffer.write((int) ((value>>62) | 1));
											}
											dataBuffer.write((int) ((value>>55) | 1));
										}
										dataBuffer.write((int) ((value>>48) | 1));
									}
									dataBuffer.write((int) ((value>>41) | 1));
								}
								dataBuffer.write((int) ((value>>34) | 1));
							}
							dataBuffer.write((int) ((value>>27) | 1));
						}
						dataBuffer.write((int) ((value>>20) | 1));
					}
					dataBuffer.write((int) ((value>>13) | 1));
				}
				dataBuffer.write((int) ((value>>6) | 1));
			}
			dataBuffer.write((int) ((value<<1) | 0));
		}
	}

	@Override
	public void writeBoolean(boolean value) {
		dataBuffer.write(value ? TRUE : FALSE);
	}

	@Override
	public void writeFloat(float value) {
		try {
			new DataOutputStream(dataBuffer).writeFloat(value);
		} catch (IOException e) {
			throw new InternalError(e.toString());
		}
	}

	@Override
	public void writeDouble(double value) {
		try {
			new DataOutputStream(dataBuffer).writeDouble(value);
		} catch (IOException e) {
			throw new InternalError(e.toString());
		}
	}

	@Override
	public void writeNLSStringId(int id){
		// Writes the ID and set the internationalized bit to 1
		writeVarUInt((id << 1) | 1);
	}

	@Override
	public void writeNonNLSString(String string){
		int address = addStringInConstantPool(string);
		// Writes the address of the string and set the internationalized bit to 0
		writeVarUInt(address<<1 | 0);
	}

	@Override
	public void markPosition(int identifier) {
		objectMap.put(Integer.valueOf(identifier), identifier, dataBuffer.getPosition());
	}

	@Override
	public void complete(long binaryId) throws IOException {
		GIMLBinaryInfo info = computeGIMLBinaryInfo(binaryId);
		writeHeader(info);
		writeObjectMap();
		writeData();
		writeConstantPool();

		out.flush();
		out.close();
		dataBuffer.close();
		constantPoolBuffer.close();
	}

	/**
	 * Computes information for the current GIML binary.
	 */
	private GIMLBinaryInfo computeGIMLBinaryInfo(long binaryId) {
		GIMLBinaryInfo info = new GIMLBinaryInfo();
		int objectMapSize = computeObjectMapSize();
		info.binaryId = binaryId;
		info.dataOffset = HEADER_SIZE + objectMapSize;
		info.constantPoolOffset = info.dataOffset + dataBuffer.getPosition();
		return info;
	}

	/**
	 * Writes GIML binary header.
	 */
	private void writeHeader(GIMLBinaryInfo info) throws IOException {
		DataOutputStream dataOut = new DataOutputStream(out);
		dataOut.writeLong(info.binaryId);
		dataOut.writeChar(info.constantPoolOffset);
		dataOut.writeChar(info.dataOffset);
	}

	/** Writes the String to offset map. */
	private void writeObjectMap() throws IOException {
		objectMap.writeTo(out);
	}

	/** Writes the objects. */
	private void writeData() throws IOException {
		dataBuffer.writeTo(out);
	}

	/** Writes the constant pool that contains the strings. */
	private void writeConstantPool() throws IOException {
		constantPoolBuffer.writeTo(out);
	}

	/**
	 * Returns the size of the object map in bytes.
	 */
	private int computeObjectMapSize() {
		return objectMap.complete();
	}

	/**
	 * Add a string in the constant pool and returns the offset of the string in the constant pool.<p>
	 * If the given string is already in the constant pool, returns the address of this string.
	 * @param string the string to add.
	 * @return the relative address of the string in the constant pool.
	 */
	private int addStringInConstantPool(String string) {
		//Check if the string is already in the constant pool
		Integer address = constantPoolEntries.get(string);
		if(address == null){
			// String is not in the constant pool: encode it and add it.
			byte[] data = stringEncoder.encodeString(string);
			address = Integer.valueOf(constantPoolBuffer.getPosition());
			constantPoolBuffer.write(data, 0, data.length);

			constantPoolEntries.put(string, address);
		}
		return address.intValue();
	}

	static class GIMLBinaryInfo {
		/** The binary identifier */
		public long binaryId;
		/** Offset of the data (the objects) in the binary */
		public int dataOffset;
		/** Offset of the constant pool in the binary */
		public int constantPoolOffset;
	}

}
