/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.cssgenerator;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringWriter;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.ServiceLoader;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.stringtemplate.v4.ST;
import org.stringtemplate.v4.STGroup;
import org.stringtemplate.v4.STGroupFile;
import org.stringtemplate.v4.StringRenderer;
import org.w3c.css.sac.InputSource;
import org.w3c.css.sac.Selector;
import org.w3c.css.sac.SelectorList;
import org.w3c.dom.css.CSSRule;
import org.w3c.dom.css.CSSRuleList;
import org.w3c.dom.css.CSSStyleDeclaration;
import org.w3c.dom.css.CSSStyleRule;
import org.w3c.dom.css.CSSStyleSheet;

import com.is2t.giml.cssgenerator.selector.SelectorFactory;
import com.is2t.giml.cssgenerator.util.CSSParserUtils;
import com.is2t.giml.description.StyleDeclaration;
import com.is2t.giml.description.StyleDescription;
import com.is2t.giml.description.StyleDescriptionFactory;
import com.steadystate.css.parser.CSSOMParser;

/**
 * The class {@code CSSGenerator} allows to generate a Java source code file containing CSS style declaration.
 *
 * @see StyleDeclaration
 * @see StyleDescription
 */
public class CSSGenerator {

	/**
	 * Default generated package name.
	 */
	public static final String DEFAULT_GENERATED_PACKAGE_NAME = "com.is2t.giml.css";

	/**
	 * Default generated class name.
	 */
	public static final String DEFAULT_GENERATED_CLASS_NAME = "CSSStyle";

	/**
	 * Default generated method name.
	 */
	public static final String DEFAULT_GENERATED_METHOD_NAME = "loadStyles";

	private static final Logger LOGGER = Logger.getLogger(CSSGenerator.class.getSimpleName());

	private static final String TEMPLATE_PATH = "com/is2t/giml/cssgenerator/template.stg";
	private static final String TEMPLATE_GROUP = "CSSGenerator";
	private static final String TEMPLATE_PACKAGE_VAR = "package";
	private static final String TEMPLATE_CLASS_VAR = "class";
	private static final String TEMPLATE_STYLE_VAR = "style";
	private static final String TEMPLATE_METHOD_VAR = "method";

	private static final String STYLE_VAR_NAME = "style";
	private static final String STYLESHEET_VAR_NAME = "styleSheet";

	private final Set<CSSStyleSheet> _styleSheets; // NOSONAR
	private String _generatedPackageName; // NOSONAR
	private String _generatedClassName; // NOSONAR
	private String _generatedMethodName; // NOSONAR
	private final List<StyleDescription> _styleDescriptions;
	private StyleDeclaration _styleDeclaration;

	/**
	 * Instantiate a new {@code CSSGenerator} with default values.
	 *
	 * <ul>
	 * <li>Default generated package name is {@value CSSGenerator#DEFAULT_GENERATED_PACKAGE_NAME}</li>
	 * <li>Default generated class name is {@value CSSGenerator#DEFAULT_GENERATED_CLASS_NAME}</li>
	 * <li>Default generated method name is {@value CSSGenerator#DEFAULT_GENERATED_METHOD_NAME}</li>
	 * </ul>
	 */
	public CSSGenerator() {
		this._styleSheets = new HashSet<CSSStyleSheet>();
		this._generatedPackageName = DEFAULT_GENERATED_PACKAGE_NAME;
		this._generatedClassName = DEFAULT_GENERATED_CLASS_NAME;
		this._generatedMethodName = DEFAULT_GENERATED_METHOD_NAME;
		this._styleDescriptions = StyleDescriptionFactory.getDescriptions();
	}

	/**
	 * Parses a CSS file and stores its rules for future generation.
	 *
	 * @param cssfile
	 *            the CSS file to parse.
	 * @return the parsed CSS file as a {@link CSSStyleSheet}.
	 * @throws IOException
	 *             if file cannot be read or something went wrong during parsing.
	 *
	 * @see #generate(File)
	 */
	public CSSStyleSheet parseFile(File cssfile) throws IOException {
		CSSOMParser parser = CSSParserUtils.getParser();
		try (FileReader fr = new FileReader(cssfile)) {
			CSSStyleSheet sheet = parser.parseStyleSheet(new InputSource(fr), null, null);
			this._styleSheets.add(sheet);
			return sheet;
		}
	}

	/**
	 * Generates all saved CSS rules into a Java source code file.
	 *
	 * @param outputDir
	 *            the output directory where the generated Java class will be generated. This directory must be a source
	 *            directory. All needed packages will be automatically generated.
	 * @throws IllegalArgumentException
	 *             if given output directory denotes a file path.
	 * @throws IOException
	 *             if something went wrong during file generation.
	 *
	 * @see #parseFile(File)
	 * @see #setGeneratedPackageName(String)
	 * @see #setGeneratedClassName(String)
	 */
	public void generate(File outputDir) throws IllegalArgumentException, IOException {
		if (outputDir.exists() && outputDir.isFile()) {
			throw new IllegalArgumentException(
					String.format("%s does not denote a directory path.", outputDir.getAbsolutePath()));
		}

		if (!outputDir.exists()) {
			outputDir.mkdirs();
		}

		// We take the first implementation of StyleDeclaration. We assume there is only one valid implem:
		// EditableStyle from ej.style
		ServiceLoader<StyleDeclaration> styleDeclarationService = ServiceLoader.load(StyleDeclaration.class);
		Iterator<StyleDeclaration> styleDeclarationIterator = styleDeclarationService.iterator();
		if (!styleDeclarationIterator.hasNext()) {
			throw new RuntimeException(String.format("Cannot find implementation of service %s",
					StyleDeclaration.class.getCanonicalName()));
		}
		this._styleDeclaration = styleDeclarationIterator.next();

		// Generate style sheet declaration
		this._styleDeclaration.generateStyleSheetDeclaration(STYLESHEET_VAR_NAME);

		// Generate style content
		for (CSSStyleSheet styleSheet : this._styleSheets) {
			this.generateStyleSheetStyle(styleSheet);
		}

		// Generate Java source code file
		this.generateStyleFile(outputDir);
	}

	/**
	 * Gets an immutable list of parsed style sheets.
	 *
	 * @return an immutable list of already parsed style sheets.
	 */
	public Set<CSSStyleSheet> getStyleSheets() {
		return Collections.unmodifiableSet(this._styleSheets);
	}

	/**
	 * Sets the generated Java source file package name.
	 *
	 * @param name
	 *            the package name.
	 *
	 * @see #DEFAULT_GENERATED_PACKAGE_NAME
	 */
	public void setGeneratedPackageName(String name) {
		this._generatedPackageName = name;
	}

	/**
	 * Sets the generated Java source file class name.
	 *
	 * @param name
	 *            the class name.
	 *
	 * @see #DEFAULT_GENERATED_CLASS_NAME
	 */
	public void setGeneratedClassName(String name) {
		this._generatedClassName = name;
	}

	/**
	 * Sets the generated Java source file method name.
	 *
	 * @param name
	 *            the method name.
	 *
	 * @see #DEFAULT_GENERATED_METHOD_NAME
	 */
	public void setGeneratedMethodName(String name) {
		this._generatedMethodName = name;
	}

	private void generateStyleSheetStyle(CSSStyleSheet styleSheet) throws IOException {
		CSSRuleList cssRules = styleSheet.getCssRules();
		for (int r = 0; r < cssRules.getLength(); r++) {
			CSSRule rule = cssRules.item(r);
			this.generateCSSRuleStyle(rule);
		}
	}

	private void generateCSSRuleStyle(CSSRule rule) throws IOException {
		StringWriter writer = this._styleDeclaration.getWriter();

		// First, read all declaration in the given CSS rule
		// Use a LinkedHashMap to preserve declaration order
		LinkedHashMap<String, String> rules = new LinkedHashMap<String, String>();
		CSSStyleRule styleRule = CSSParserUtils.getRule(rule, CSSRule.STYLE_RULE);
		if (styleRule == null) {
			LOGGER.log(Level.WARNING,
					String.format("Unsupported CSS rule. It will be ignored. (found %s)", rule.toString()));
			return;
		}
		CSSStyleDeclaration cssStyleDeclaration = styleRule.getStyle();
		for (int s = 0; s < cssStyleDeclaration.getLength(); s++) {
			String ruleDeclaration = cssStyleDeclaration.item(s);
			rules.put(ruleDeclaration, cssStyleDeclaration.getPropertyValue(ruleDeclaration));
		}

		// Append to generated file the original CSS rule for readability
		writer.append("/*\n\t ").append(rule.toString()).append("\n*/\n");

		// Generate content of the given rule into the style declaration
		this._styleDeclaration.generateStyleInitialization(STYLE_VAR_NAME);
		for (StyleDescription styleDescription : this._styleDescriptions) {
			styleDescription.generateStyle(STYLE_VAR_NAME, rules, writer);
		}
		SelectorList selectors = CSSParserUtils.getSelectors(styleRule.getSelectorText());
		for (int i = 0; i < selectors.getLength(); i++) {
			Selector selector = selectors.item(i);
			String selectorVar = SelectorFactory.create(selector, writer);
			writer.append(STYLESHEET_VAR_NAME).append(".addRule(");
			writer.append(selectorVar);
			writer.append(", ").append(STYLE_VAR_NAME).append(");\n");
		}
		writer.append("\n");
	}

	private void generateStyleFile(File outputDir) throws IOException {
		STGroup stGroup = new STGroupFile(TEMPLATE_PATH);
		stGroup.registerRenderer(String.class, new StringRenderer());
		ST template = stGroup.getInstanceOf(TEMPLATE_GROUP);
		template.add(TEMPLATE_PACKAGE_VAR, this._generatedPackageName);
		template.add(TEMPLATE_CLASS_VAR, this._generatedClassName);
		template.add(TEMPLATE_METHOD_VAR, this._generatedMethodName);
		template.add(TEMPLATE_STYLE_VAR, this._styleDeclaration.getWriter().toString());
		File generatedFile = new File(outputDir, this.getGeneratedFileRelativePath());
		generatedFile.getParentFile().mkdirs();
		generatedFile.createNewFile();
		template.write(generatedFile, null);
	}

	private String getGeneratedFileRelativePath() {
		return this._generatedPackageName.replace('.', '/') + "/" + this._generatedClassName + ".java";
	}
}
