/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.generator.test.parsergenerator;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.util.Set;

import org.apache.commons.io.FileUtils;
import org.junit.Assert;
import org.junit.Test;

import com.is2t.giml.xml.generator.XMLGenerator;
import com.is2t.giml.xml.generator.XMLGeneratorOptions;
import com.is2t.giml.xml.generator.exception.XMLGeneratorException;
import com.is2t.giml.xml.generator.test.AbstractTest;
import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.Document;
import com.is2t.giml.xml.model.Node;
import com.is2t.giml.xml.parser.XMLParser;
import com.is2t.giml.xml.parser.XMLParserOptions;
import com.is2t.giml.xml.parser.exception.XMLParserException;

/**
 * Test cases of {@link XMLParser} on valid documents and resolved imports.
 */
public class XMLParserGeneratorTest extends AbstractTest {

	private static final String RESOURCE_FOLDER_PATH = "/com/is2t/giml/xml/generator/test/parsergenerator/";

	private static final String PARSER_OPTION_XML_EXTENSION = ".xml";
	private static final String PARSER_OPTION_DOCUMENT_ID_ATTRIBUTE_NAME = "name";
	private static final String PARSER_OPTION_IMPORT_TAGNAME = "import";
	private static final String PARSER_OPTION_IMPORT_ATTRIBUTE_NAME = "name";
	private static final boolean PARSER_OPTION_HANDLE_IMPORTS = true;

	private static final String PARSER_OPTION_TEMPLATE_CONTENT_TAG = "content";
	private static final String PARSER_OPTION_TEMPLATE_CONTENT_TAG_INSTANTIATION_ATTRIBUTE_NAME = "ref";

	/**
	 * Gets a {@link File} instance on a resource in the current JAR.
	 *
	 * @param resourceName
	 *            the name of the resource to get, relative to the resource folder path of this test.
	 * @return the file in the JAR.
	 *
	 * @see #RESOURCE_FOLDER_PATH
	 */
	protected File getFileFromJarResource(String resourceName) {
		String resourcePath = this.getResourceFolder() + resourceName;
		URL fileURL = AbstractTest.class.getResource(resourcePath);
		return FileUtils.toFile(fileURL);
	}

	/**
	 * Gets the XML parser instance to use for tests.
	 *
	 * @param rootFile
	 *            the root file for the parser.
	 * @return an XML parser.
	 * @throws FileNotFoundException
	 *             if the given file does not denote an existing file.
	 */
	protected XMLParser createXMLParser(File rootFile) throws FileNotFoundException {
		XMLParser parser = new XMLParser(rootFile);
		parser.setOptions(this.getXMLParserOptions());
		return parser;
	}

	/**
	 * Gets the XML generator instance to use for tests.
	 *
	 * @param document
	 *            the root document for the generator.
	 * @return an XML generator.
	 */
	protected XMLGenerator createXMLGenerator(Document document) {
		XMLGenerator generator = new XMLGenerator(document);
		generator.setOptions(this.getXMLGeneratorOptions());
		return generator;
	}

	/**
	 * Gets the {@link XMLParserOptions} for the test parser.
	 *
	 * @return options for the test parser.
	 */
	protected XMLParserOptions getXMLParserOptions() {
		XMLParserOptions options = new XMLParserOptions();
		options.setXMLFileExtension(PARSER_OPTION_XML_EXTENSION);
		options.setDocumentIdAttributeName(PARSER_OPTION_DOCUMENT_ID_ATTRIBUTE_NAME);
		options.setImportTagName(PARSER_OPTION_IMPORT_TAGNAME);
		options.setImportTagAttributeName(PARSER_OPTION_IMPORT_ATTRIBUTE_NAME);
		options.setHandleImports(PARSER_OPTION_HANDLE_IMPORTS);
		return options;
	}

	/**
	 * Gets the {@link XMLGeneratorOptions} for the test generator.
	 *
	 * @return options for the test generator.
	 */
	protected XMLGeneratorOptions getXMLGeneratorOptions() {
		XMLGeneratorOptions options = new XMLGeneratorOptions();
		options.setXMLFileExtension(PARSER_OPTION_XML_EXTENSION);
		options.setDocumentIdAttributeName(PARSER_OPTION_DOCUMENT_ID_ATTRIBUTE_NAME);
		options.setImportTagName(PARSER_OPTION_IMPORT_TAGNAME);
		options.setImportTagAttributeName(PARSER_OPTION_IMPORT_ATTRIBUTE_NAME);
		return options;
	}

	/**
	 * Gets the resource folder for tests.
	 *
	 * @return he resource folder path for tests.
	 */
	protected String getResourceFolder() {
		return RESOURCE_FOLDER_PATH;
	}

	@Test
	public void testSimpleDocument() throws XMLParserException, IOException, XMLGeneratorException {
		Document document = createXMLParser(getFileFromJarResource("simple.xml")).parse();

		File tempDir = createTempDir("simple");
		System.out.println(tempDir.getAbsolutePath());
		File tempFile = new File(tempDir, "simple.xml");
		tempFile.createNewFile();
		createXMLGenerator(document).generate(tempFile);
		Document document2 = createXMLParser(tempFile).parse();

		Assert.assertEquals(document, document2);
	}

	@Test
	public void testDocumentsWithChainImport() throws XMLParserException, IOException, XMLGeneratorException {
		Document document = createXMLParser(getFileFromJarResource("importdoc1.xml")).parse();

		File tempDir = createTempDir("import");
		System.out.println(tempDir.getAbsolutePath());
		createXMLGenerator(document).generateRecursively(tempDir);
		Document document2 = createXMLParser(new File(tempDir, "doc1.xml")).parse();

		Assert.assertEquals(document, document2);

		for (int i = 1; ++i <= 4;) {
			Document reference = createXMLParser(getFileFromJarResource("importdoc" + i + ".xml")).parse();
			Document generated = createXMLParser(new File(tempDir, "doc" + i + ".xml")).parse();
			Assert.assertEquals(reference, generated);
		}
	}

	@Test
	public void testMultipleTemplatesWithChildrens() throws XMLParserException, IOException, XMLGeneratorException {
		Document document = createXMLParser(getFileFromJarResource("templatemultiple.xml")).parse();

		File tempDir = createTempDir("templatemultiple");
		File tempFile = new File(tempDir, "templatemultiple.xml");
		tempFile.createNewFile();
		System.out.println(tempDir.getAbsolutePath());
		createXMLGenerator(document).generate(tempFile);
		Document document2 = createXMLParser(tempFile).parse();

		Assert.assertEquals(document, document2);

		Assert.assertNotNull(document);
		Assert.assertEquals(document.getName(), "root");
		Assert.assertEquals(document.getId(), "multiple");
		Assert.assertEquals(document.getAttributeCount(), 1); // the ID is an attribute
		Assert.assertEquals(document.getReferencedDocumentCount(), 0);

		// Check before resolving template
		Set<Attribute> invocationNode1Attributes = createAttributeSet("name", "John");
		Node invocationNode1 = createNode("say-hello", invocationNode1Attributes);
		Node invocationNode2Child = createNode(PARSER_OPTION_TEMPLATE_CONTENT_TAG,
				createAttributeSet(PARSER_OPTION_TEMPLATE_CONTENT_TAG_INSTANTIATION_ATTRIBUTE_NAME, "myContent"),
				createNode("node2", NO_ATTRIBUTE));
		Set<Attribute> invocationNode2Attributes = createAttributeSet("firstname", "John", "lastname", "Doe");
		Node invocationNode2 = createNode("say-goodbye", invocationNode2Attributes, invocationNode2Child);
		Node container = createNode("node", NO_ATTRIBUTE, invocationNode1, invocationNode2);

		Assert.assertEquals(document.getChildCount(), 1);
		Assert.assertEquals(document.getTemplateCount(), 2);
		Assert.assertTrue(document.getChildren().contains(container));

		// Check after resolving template
		document = document.resolveTemplates();
		Node expectedNode = createNode("node", NO_ATTRIBUTE,
				createNode("node1", createAttributeSet("text", "Hello John!")), createNode("node1",
						createAttributeSet("text", "Goodbye John Doe!"), createNode("node2", NO_ATTRIBUTE)));
		Assert.assertEquals(document.getChildCount(), 1);
		Assert.assertEquals(document.getTemplateCount(), 0);
		Assert.assertTrue(document.getChildren().contains(expectedNode));
	}

}
