/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.parser;

import java.util.Objects;

/**
 * The class {@code XMLParserOptions} represents the configuration of a {@link XMLParser} instance. Root tag, import tag
 * and other options are configurable.
 */
public class XMLParserOptions {

	private String _xmlFileExtension; // NOSONAR

	private String _documentIdAttributeName; // NOSONAR

	private boolean _handleImports; // NOSONAR
	private String _importTagName; // NOSONAR
	private String _importTagAttributeName; // NOSONAR

	private boolean _handleTemplates; // NOSONAR
	private String _templateTagName; // NOSONAR
	private String _templateTagAttributeName; // NOSONAR
	private String _templateParameterPrefix; // NOSONAR
	private String _templateParameterSuffix; // NOSONAR
	private String _templateContentTagName; // NOSONAR
	private String _templateContentAttributeName; // NOSONAR
	private String _templateContentInstantiationAttributeName; // NOSONAR

	/**
	 * Create new options for an {@link XMLParser} instance, with default values:
	 * <ul>
	 * <li>Default handled XML files have extension
	 * {@value com.is2t.giml.xml.parser.XMLParserConstants#DEFAULT_XML_FILE_EXTENSION}</li>
	 * <li>Default document ID attribute name is
	 * {@value com.is2t.giml.xml.parser.XMLParserConstants#DEFAULT_DOCUMENT_ID_ATTRIBUTE_NAME}</li>
	 * <li>Default import tag name is
	 * {@value com.is2t.giml.xml.parser.XMLParserConstants#DEFAULT_IMPORT_DOCUMENT_TAGNAME}</li>
	 * <li>Default import name attribute name is
	 * {@value com.is2t.giml.xml.parser.XMLParserConstants#DEFAULT_IMPORT_NAME_ATTRIBUTE_NAME}</li>
	 * <li>Default template tag name is {@value com.is2t.giml.xml.parser.XMLParserConstants#DEFAULT_TEMPLATE_TAGNAME}
	 * </li>
	 * <li>Default template name attribute name is
	 * {@value com.is2t.giml.xml.parser.XMLParserConstants#DEFAULT_TEMPLATE_NAME_ATTRIBUTE_NAME}</li>
	 * <li>Default template name attribute name is
	 * {@value com.is2t.giml.xml.parser.XMLParserConstants#DEFAULT_TEMPLATE_NAME_ATTRIBUTE_NAME}</li>
	 * <li>Default template parameter prefix is
	 * {@value com.is2t.giml.xml.parser.XMLParserConstants#DEFAULT_TEMPLATE_PARAMETER_PREFIX}</li>
	 * <li>Default template parameter suffix is
	 * {@value com.is2t.giml.xml.parser.XMLParserConstants#DEFAULT_TEMPLATE_PARAMETER_SUFFIX}</li>
	 * <li>Default template content tag name is
	 * {@value com.is2t.giml.xml.parser.XMLParserConstants#DEFAULT_TEMPLATE_CONTENT_TAGNAME}</li>
	 * <li>Default template content name attribute name is
	 * {@value com.is2t.giml.xml.parser.XMLParserConstants#DEFAULT_TEMPLATE_CONTENT_NAME_ATTRIBUTE_NAME}</li>
	 * <li>Default template instantiation content name attribute name is
	 * {@value com.is2t.giml.xml.parser.XMLParserConstants#DEFAULT_TEMPLATE_CONTENT_INSTANTIATION_NAME_ATTRIBUTE_NAME}
	 * </li>
	 * </ul>
	 *
	 * By default, parser import support is {@value com.is2t.giml.xml.parser.XMLParserConstants#DEFAULT_HANDLE_IMPORTS}
	 * and template support is {@value com.is2t.giml.xml.parser.XMLParserConstants#DEFAULT_HANDLE_TEMPLATES}.
	 */
	public XMLParserOptions() {
		this.setDefaultOptions();
	}

	/**
	 * Gets file extension of files handled by the parser.
	 *
	 * @return the extension of files handled by the parser.
	 */
	public String getXMLFileExtension() {
		return this._xmlFileExtension;
	}

	/**
	 * Gets the attribute name that describe document unique ID in document root tag.
	 *
	 * @return the attribute name that describe document unique ID.
	 */
	public String getDocumentIdAttributeName() {
		return this._documentIdAttributeName;
	}

	/**
	 * Gets if the parser must handle imports or not.
	 *
	 * @return <code>true</code> if parser must handle imports, <code>false</code> otherwise.
	 */
	public boolean handleImports() {
		return this._handleImports;
	}

	/**
	 * Gets the tag name describing document imports.
	 *
	 * @return the tag name describing document imports.
	 */
	public String getImportTagName() {
		return this._importTagName;
	}

	/**
	 * Gets the attribute name that describe the document to import in import tags.
	 *
	 * @return the attribute name that describe the document to import.
	 */
	public String getImportTagAttributeName() {
		return this._importTagAttributeName;
	}

	/**
	 * Gets if the parser must handle templates or not.
	 *
	 * @return <code>true</code> if parser must handle templates, <code>false</code> otherwise.
	 */
	public boolean handleTemplates() {
		return this._handleTemplates;
	}

	/**
	 * Gets the tag name describing templates.
	 *
	 * @return the tag name describing templates.
	 */
	public String getTemplateTagName() {
		return this._templateTagName;
	}

	/**
	 * Gets the attribute name that describe the template name in template tags.
	 *
	 * @return the attribute name that describe the template name in template tags.
	 */
	public String getTemplateTagAttributeName() {
		return this._templateTagAttributeName;
	}

	/**
	 * Gets the prefix that identify the start of a template parameter.
	 *
	 * @return the prefix that identify the start of a template parameter.
	 */
	public String getTemplateParameterPrefix() {
		return this._templateParameterPrefix;
	}

	/**
	 * Gets the suffix that identify the end of a template parameter.
	 *
	 * @return the suffix that identify the end of a template parameter.
	 */
	public String getTemplateParameterSuffix() {
		return this._templateParameterSuffix;
	}

	/**
	 * Gets the tag name describing template contents.
	 *
	 * @return the tag name describing template contents.
	 */
	public String getTemplateContentTagName() {
		return this._templateContentTagName;
	}

	/**
	 * Gets the attribute name that describe the template content name in template content tags.
	 *
	 * @return the attribute name that describe the template content name in template content tags.
	 */
	public String getTemplateContentTagAttributeName() {
		return this._templateContentAttributeName;
	}

	/**
	 * Gets the attribute name that describe the template instantiation content name in template content tags.
	 *
	 * @return the attribute name that describe the template instantiation content name in template content tags.
	 */
	public String getTemplateContentTagInstantiationAttributeName() {
		return this._templateContentInstantiationAttributeName;
	}

	/**
	 * Sets file extension for files handler by the parser.
	 *
	 * @param extension
	 *            the extension to look for files to parse.
	 *
	 * @throws IllegalArgumentException
	 *             if given value is <code>null</code>, empty or contains spaces.
	 */
	public void setXMLFileExtension(String extension) throws IllegalArgumentException {
		checkNotNullOrEmptyOrSpaces("file extension", extension);
		this._xmlFileExtension = extension;
	}

	/**
	 * Sets attribute name that describe document unique ID in document root tags.
	 *
	 * @param attributeName
	 *            the name of the attribute that describe the document unique ID.
	 *
	 * @throws IllegalArgumentException
	 *             if given value is <code>null</code> or empty.
	 */
	public void setDocumentIdAttributeName(String attributeName) throws IllegalArgumentException {
		checkNotNullOrEmptyOrSpaces("document id attribute name", attributeName);
		this._documentIdAttributeName = attributeName;
	}

	/**
	 * Sets if the parser must handle imports tags. Import tags can be customized.
	 *
	 * @param value
	 *            <code>true</code> if the parser must process import tags, <code>false</code> otherwise.
	 */
	public void setHandleImports(boolean value) {
		this._handleImports = value;
	}

	/**
	 * Sets tag name for import declarations.
	 *
	 * @param tagname
	 *            the tag name for importing documents.
	 *
	 * @throws IllegalArgumentException
	 *             if given value is <code>null</code> or empty.
	 */
	public void setImportTagName(String tagname) throws IllegalArgumentException {
		checkNotNullOrEmptyOrSpaces("import tag name", tagname);
		this._importTagName = tagname;
	}

	/**
	 * Sets attribute name that describe the document to import in import tags.
	 *
	 * @param attributeName
	 *            the name of the attribute that describe the document to import.
	 *
	 * @throws IllegalArgumentException
	 *             if given value is <code>null</code> or empty.
	 */
	public void setImportTagAttributeName(String attributeName) throws IllegalArgumentException {
		checkNotNullOrEmptyOrSpaces("import tag attribute name", attributeName);
		this._importTagAttributeName = attributeName;
	}

	/**
	 * Sets if the parser must handle template tags. Template tags can be customized.
	 *
	 * @param value
	 *            <code>true</code> if the parser must process template tags, <code>false</code> otherwise.
	 */
	public void setHandleTemplates(boolean value) {
		this._handleTemplates = value;
	}

	/**
	 * Sets tag name for template declarations.
	 *
	 * @param tagname
	 *            the tag name to declare templates.
	 *
	 * @throws IllegalArgumentException
	 *             if given value is <code>null</code> or empty.
	 */
	public void setTemplateTagName(String tagname) throws IllegalArgumentException {
		checkNotNullOrEmptyOrSpaces("template tag name", tagname);
		this._templateTagName = tagname;
	}

	/**
	 * Sets attribute name that describe the name of the template to declare.
	 *
	 * @param attributeName
	 *            the name of the attribute that describe the name of the template to declare.
	 *
	 * @throws IllegalArgumentException
	 *             if given value is <code>null</code> or empty.
	 */
	public void setTemplateTagAttributeName(String attributeName) throws IllegalArgumentException {
		checkNotNullOrEmptyOrSpaces("template tag attribute name", attributeName);
		this._templateTagAttributeName = attributeName;
	}

	/**
	 * Sets prefix that starts template parameters.
	 *
	 * @param prefix
	 *            the prefix to use to determine template parameters.
	 *
	 * @throws IllegalArgumentException
	 *             if given value is <code>null</code> or empty.
	 */
	public void setTemplateParameterPrefix(String prefix) throws IllegalArgumentException {
		checkNotNullOrEmptyOrSpaces("template parameter prefix", prefix);
		this._templateParameterPrefix = prefix;
	}

	/**
	 * Sets suffix that ends template parameters.
	 *
	 * @param suffix
	 *            the suffix to use to determine template parameters.
	 *
	 * @throws IllegalArgumentException
	 *             if given value is <code>null</code> or empty.
	 */
	public void setTemplateParameterSuffix(String suffix) throws IllegalArgumentException {
		checkNotNullOrEmptyOrSpaces("template parameter suffix", suffix);
		this._templateParameterSuffix = suffix;
	}

	/**
	 * Sets tag name for template content declarations.
	 *
	 * @param tagname
	 *            the tag name to declare template contents.
	 *
	 * @throws IllegalArgumentException
	 *             if given value is <code>null</code> or empty.
	 */
	public void setTemplateContentTagName(String tagname) throws IllegalArgumentException {
		checkNotNullOrEmptyOrSpaces("template content tag name", tagname);
		this._templateContentTagName = tagname;
	}

	/**
	 * Sets attribute name that describe the name of the template content to declare.
	 *
	 * @param attributeName
	 *            the name of the attribute that describe the name of the template content to declare.
	 *
	 * @throws IllegalArgumentException
	 *             if given value is <code>null</code> or empty.
	 */
	public void setTemplateContentTagAttributeName(String attributeName) throws IllegalArgumentException {
		checkNotNullOrEmptyOrSpaces("template content tag attribute name", attributeName);
		this._templateContentAttributeName = attributeName;
	}

	/**
	 * Sets attribute name that describe the name of the template instantiation content to declare.
	 *
	 * @param attributeName
	 *            the name of the attribute that describe the name of the template instantiation content to declare.
	 *
	 * @throws IllegalArgumentException
	 *             if given value is <code>null</code> or empty.
	 */
	public void setTemplateContentTagInstantiationAttributeName(String attributeName) throws IllegalArgumentException {
		checkNotNullOrEmptyOrSpaces("template instantiation content tag attribute name", attributeName);
		this._templateContentInstantiationAttributeName = attributeName;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}

		if (!(obj instanceof XMLParserOptions)) {
			return false;
		}

		XMLParserOptions other = (XMLParserOptions) obj;
		return this._xmlFileExtension.equals(other._xmlFileExtension) && this.areImportOptionsEqual(other)
				&& this._documentIdAttributeName.equals(other._documentIdAttributeName)
				&& this.areTemplateOptionsEqual(other);
	}

	@Override
	public int hashCode() {
		return (this._handleImports ? 1 : 0) + (this._handleTemplates ? 1 : 0)
				+ Objects.hash(this._xmlFileExtension, this._importTagName, this._importTagAttributeName,
						this._documentIdAttributeName, this._templateTagName, this._templateTagAttributeName,
						this._templateParameterPrefix, this._templateParameterSuffix);
	}

	private void setDefaultOptions() {
		this._xmlFileExtension = XMLParserConstants.DEFAULT_XML_FILE_EXTENSION;

		this._documentIdAttributeName = XMLParserConstants.DEFAULT_DOCUMENT_ID_ATTRIBUTE_NAME;

		this._handleImports = XMLParserConstants.DEFAULT_HANDLE_IMPORTS;
		this._importTagName = XMLParserConstants.DEFAULT_IMPORT_DOCUMENT_TAGNAME;
		this._importTagAttributeName = XMLParserConstants.DEFAULT_IMPORT_NAME_ATTRIBUTE_NAME;

		this._handleTemplates = XMLParserConstants.DEFAULT_HANDLE_TEMPLATES;
		this._templateTagName = XMLParserConstants.DEFAULT_TEMPLATE_TAGNAME;
		this._templateTagAttributeName = XMLParserConstants.DEFAULT_TEMPLATE_NAME_ATTRIBUTE_NAME;
		this._templateParameterPrefix = XMLParserConstants.DEFAULT_TEMPLATE_PARAMETER_PREFIX;
		this._templateParameterSuffix = XMLParserConstants.DEFAULT_TEMPLATE_PARAMETER_SUFFIX;
		this._templateContentTagName = XMLParserConstants.DEFAULT_TEMPLATE_CONTENT_TAGNAME;
		this._templateContentAttributeName = XMLParserConstants.DEFAULT_TEMPLATE_CONTENT_NAME_ATTRIBUTE_NAME;
		this._templateContentInstantiationAttributeName = XMLParserConstants.DEFAULT_TEMPLATE_CONTENT_INSTANTIATION_NAME_ATTRIBUTE_NAME;
	}

	private void checkNotNullOrEmptyOrSpaces(String context, String value) throws IllegalArgumentException {
		if (value == null) {
			throw new IllegalArgumentException("Value cannot be null for " + context);
		}

		if (value.isEmpty()) {
			throw new IllegalArgumentException("Value cannot be empty for " + context);
		}
		if (value.matches(".*\\s.*")) {
			throw new IllegalArgumentException("Value cannot contains spaces for " + context);
		}
	}

	private boolean areImportOptionsEqual(XMLParserOptions other) {
		return this._importTagName.equalsIgnoreCase(other._importTagName)
				&& this._importTagAttributeName.equalsIgnoreCase(other._importTagAttributeName)
				&& this._handleImports == other._handleImports;
	}

	private boolean areTemplateOptionsEqual(XMLParserOptions other) {
		return this._templateTagName.equalsIgnoreCase(other._templateTagName)
				&& this._templateTagAttributeName.equalsIgnoreCase(other._templateTagAttributeName)
				&& this._templateParameterPrefix.equals(other._templateParameterPrefix)
				&& this._templateParameterSuffix.equals(other._templateParameterSuffix)
				&& this.areTemplateContentOptionsEqual(other);
	}

	private boolean areTemplateContentOptionsEqual(XMLParserOptions other) {
		return this._templateContentTagName.equalsIgnoreCase(other._templateContentTagName)
				&& this._templateContentAttributeName.equalsIgnoreCase(other._templateContentAttributeName)
				&& this._templateContentInstantiationAttributeName
						.equalsIgnoreCase(other._templateContentInstantiationAttributeName);
	}
}
