/*
 * Java
 *
 * Copyright 2009-2014 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.hil;


/**
 * <p>
 * This class provides some utilities to implement simulated natives.
 * <ul>
 * <li>It allows to refresh/flush the content of arrays passed as arguments through HIL protocol. Synchronization is
 * done manually to limit byte transfer (speed optimization). An immutable array can be synchronized (refresh/flush) at
 * any time. An array passed as argument is local to the invoke context (new instance). It cannot be refreshed but it
 * may be flushed when the method that has created this argument returns.</li>
 * <li>It provides interaction to get back resources content</li>
 * <li>It also gives a framework to interact with the shutdown process of the whole simulation thanks to
 * {@link StopListener} and the {@link #stop()} method</li>
 * </ul>
 * </p>
 */
public interface NativeInterface {

	/**
	 * Tells whether an object is an immortal array or not.
	 *
	 * @param obj
	 *            the object to test.
	 * @return <code>true</code> if the given object is an immortal array, <code>false</code> otherwise.
	 */
	boolean isImmortal(Object obj);

	/**
	 * Refresh full content of array if it is immortal.<br>
	 * Does nothing if the given object is not an immortal array.<br>
	 * Equivalent to call {@link #refreshImmortalContent(Object)} assuming that the array is immortal.
	 *
	 * @see #isImmortal(Object)
	 */
	void refreshContent(Object array);

	/**
	 * Refresh content of array if it is immortal.<br>
	 * Does nothing if the given object is not an immortal array. Equivalent to call
	 * {@link #refreshImmortalContent(Object, int, int)} assuming that the array is immortal.
	 *
	 * @see #refreshImmortalContent(Object, int, int)
	 * @see #isImmortal(Object)
	 */
	void refreshContent(Object array, int offset, int length);

	/**
	 * Refresh fully an immortal array.<br>
	 * Equivalent to call {@link #refreshImmortalContent(Object, int, int)} with the full array range.
	 */
	void refreshImmortalContent(Object array);

	/**
	 * Refresh content of an immortal array.<br>
	 * Only range <code>[offset, offset + length - 1]</code> is refreshed.
	 *
	 * @param array
	 *            the immortal array to refresh.
	 * @param offset
	 *            the start index of the range.
	 * @param length
	 *            the length of the range.
	 * @throws IllegalArgumentException
	 *             if the given array is not an array or is not immortal.
	 * @throws ArrayIndexOutOfBoundsException
	 *             if the given range is invalid.
	 * @throws BrokenConnection
	 *             if the array cannot be refreshed.
	 */
	void refreshImmortalContent(Object array, int offset, int length);

	/**
	 * Flush fully an array.<br>
	 * Equivalent to call {@link #flushContent(Object, int, int)} with the full array range.
	 */
	void flushContent(Object array);

	/**
	 * Flush the content of an array.<br>
	 * Only range <code>[offset, offset + length - 1]</code> is flushed. If array is immortal, this method calls
	 * {@link #flushImmortalContent(Object, int, int)}. Otherwise, if array has been passed as argument of an invoke
	 * method, the content of the array is marked to be flushed when method returns. This method may be called many
	 * times to extend the range to flush (union of intervals).
	 *
	 * @param array
	 *            the immortal array to flush.
	 * @param offset
	 *            the start index of the range.
	 * @param length
	 *            the length of the range.
	 * @throws IllegalArgumentException
	 *             if the given array is not an array or is not immortal.
	 * @throws ArrayIndexOutOfBoundsException
	 *             if the given range is invalid.
	 * @throws BrokenConnection
	 *             if the array cannot be flushed.
	 */
	void flushContent(Object array, int offset, int length);

	/**
	 * Flush fully an immortal array.<br>
	 * Equivalent to call {@link #flushImmortalContent(Object, int, int)} with the full array range.
	 */
	void flushImmortalContent(Object array);

	/**
	 * Flush content of an immortal array.<br>
	 * Only range <code>[offset, offset + length - 1]</code> is flushed.
	 *
	 * @param array
	 *            the immortal array to flush.
	 * @param offset
	 *            the start index of the range.
	 * @param length
	 *            the length of the range.
	 * @throws IllegalArgumentException
	 *             if the given array is not an array or is not immortal.
	 * @throws ArrayIndexOutOfBoundsException
	 *             if the given range is invalid.
	 * @throws BrokenConnection
	 *             if the array cannot be flushed.
	 */
	void flushImmortalContent(Object array, int offset, int length);

	/**
	 * Get full content of a resource.
	 *
	 * @param resourceName
	 *            as described in {@link Class#getResourceAsStream(String)}.
	 * @return the content of the resource (byte array) or <code>null</code> if not found.
	 * @throws IllegalArgumentException
	 *             if the given resource name is not absolute.
	 * @throws BrokenConnection
	 *             if the resource could not be retrieved.
	 */
	byte[] getResourceContent(String resourceName);

	/**
	 * Register a stop listener.<br>
	 * The stop listeners will be notified when the simulation stops.
	 *
	 * @param stopListener
	 *            the listener to register.
	 */
	void addStopListener(StopListener stopListener);

	/**
	 * Allows to request the simulation stop.
	 */
	void stop();

	/**
	 * Gets the ID of the current Java thread.
	 *
	 * @return the ID of the current Java thread.
	 */
	int getCurrentJavaThreadID();

	/**
	 * <p>
	 * Causes the current Java thread to pause its Java execution after the end of the current native method. This
	 * function is not blocking. The current Java thread will resume its execution after the reception of an external
	 * event or after <code>timeout</code> milliseconds.
	 * </p>
	 * <p>
	 * If a resume has been done on this thread before calling this function, the thread is not paused.
	 * </p>
	 * <p>
	 * The result of calling this method several times during the same native execution is unpredictable.
	 * </p>
	 * 
	 * @param timeout
	 *            the duration in milliseconds of the pause. If it is equals to zero, then time is not taken into
	 *            consideration and the thread simply waits until resumed.
	 * @return <code>true</code> if the current java thread will suspend its execution, <code>false</code> if a resume
	 *         is pending.
	 * @throws IllegalStateException
	 *             if the method is not called inside a Java native.
	 * @see #resumeJavaThread(int)
	 */
	boolean suspendCurrentJavaThread(long timeout);

	/**
	 * <p>
	 * Resume the given Java thread if it is suspended. If the Java thread is not paused, this resume stays pending.
	 * Next call of {@link #suspendCurrentJavaThread(long)} will not paused the thread.
	 * </p>
	 * 
	 * @param javaThreadID
	 *            the ID of the Java thread to resume.
	 * @throws IllegalArgumentException
	 *             if the given Java thread ID is invalid.
	 * @see #getCurrentJavaThreadID()
	 */
	void resumeJavaThread(int javaThreadID);
	
	/**
	 * Transforms a Java String into a C String.<br>
	 * The platform default encoding is used to transform Java characters into C characters.<br>
	 * The created C String is a NULL terminated String (ends with '\0'). The <code>cString</code> array length
	 * must be at least <code>javaString.length()+1</code>.<br>
	 * The content of the given <code>cString</code> array is automatically flushed with the method {@link NativeInterface#flushContent(Object, int, int)}.
	 * Only <code>javaString.length()</code> bytes are flushed.
	 * 
	 * @param javaString the Java String
	 * @param cString byte array which contains the C String.
	 * 
	 * @throws IllegalArgumentException if javaString or cString is null
	 * @throws ArrayIndexOutOfBoundsException if cString is too small to contain the string
	 */
	void toCString(String javaString, byte[] cString);
	
	/**
	 * Transforms a C String into a Java String,
	 * using platform default encoding.
	 * The C String must be NULL terminated.<br>
	 * The content of the given <code>cString</code> is automatically refreshed with the method {@link NativeInterface#refreshContent(Object)}.
	 * 
	 * @param cString byte array which contains the C String
	 * @return a new Java String.
	 * 
	 * @throws IllegalArgumentException if cString is null or its length is < 1
	 * @throws IllegalArgumentException if cString is not NULL terminated.
	 */
	String toJavaString(byte[] cString);

}
