/*
 * Java
 * 
 * Copyright 2013 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.linker.map;

import com.is2t.tools.graph.IGraph;

/**
 * <p>
 * Abstraction of a Memory Map.
 * </p>
 * <p>
 * A Memory Map is compound of symbols and ranges.
 * </p>
 */
public interface IMemoryMap {

	/**
	 * <p>
	 * From an id (abstract notion implementation dependent), returns the name of the input unit (object file).
	 * </p>
	 * 
	 * @param id an abstraction of an input unit identifier
	 * @return the name of the input unit associated to the given id.
	 */
	String getInputUnit(String id);

	/**
	 * <p>
	 * Returns the list of symbols referenced by this memory map.
	 * </p>
	 * 
	 * @return the list of symbols referenced by this memory map.
	 */
	ISymbol[] getSymbols();

	/**
	 * <p>
	 * Returns the list of symbols included by this section.
	 * </p>
	 * 
	 * @param section the section where to find symbols
	 * @return the list of symbols included by this section.
	 */
	ISymbol[] getSymbols(IRange section);

	/**
	 * <p>
	 * Same as {@link #getSymbols(IRange)} using {@link #getSection(String)}
	 * </p>
	 */
	ISymbol[] getSymbols(String sectionName);

	/**
	 * <p>
	 * Returns the list of names of the symbols referenced by this memory map.
	 * </p>
	 * 
	 * @return the list of names of the symbols referenced by this memory map.
	 * @see #getSymbols()
	 */
	String[] getSymbolNames();

	/**
	 * <p>
	 * Returns the symbol abstraction from its name.
	 * </p>
	 * 
	 * @param name the name of the symbol
	 * @return the symbol abstraction from its name.
	 */
	ISymbol getSymbol(String name);

	/**
	 * <p>
	 * Same as {@link #findAccessPaths(ISymbol, boolean)} using {@link #getSymbol(String)}.
	 * </p>
	 */
	IGraph findAccessPaths(String symbolName, boolean stopAtFirstResult);

	/**
	 * <p>
	 * Find paths (or at least the first found one) allowing to reach the given symbol from roots.
	 * </p>
	 * 
	 * @param symbol the symbol to reach
	 * @param stopAtFirstResult stop the search at first match
	 * @return paths (or at least the first found one) allowing to reach the given symbol from roots.
	 */
	IGraph findAccessPaths(ISymbol symbol, boolean stopAtFirstResult);

	/**
	 * <p>
	 * Returns the list of the sections of this memory map.
	 * </p>
	 * 
	 * @return Returns the list of the sections of this memory map.
	 */
	IRange[] getSections();

	/**
	 * <p>
	 * Returns the list of names of the sections of this memory map.
	 * </p>
	 * 
	 * @return Returns the list of names of the sections of this memory map.
	 * @see #getSections()
	 */
	String[] getSectionNames();

	/**
	 * <p>
	 * Returns the section abstraction from its name.
	 * </p>
	 * 
	 * @param name the name of the section
	 * @return the section abstraction from its name.
	 */
	IRange getSection(String name);
}