/*
 * Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2010-2021, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.is2t.support.net;

import java.io.IOException;
import java.net.BindException;
import java.net.Inet4Address;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.net.SocketOptions;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
import java.util.Enumeration;

import com.is2t.errors.NetErrors;
import com.is2t.support.net.natives.ChannelNatives;
import com.is2t.support.net.util.IPAddressConstants;

import ej.error.Message;

/**
 * Native interface to support configuring of channel to run in a non-blocking
 * manner and support scatter/gather io operations.
 */
public class Channel implements IChannel {

	/**
	 * Our reference implementation uses an integer to store the native file
	 * descriptor. Implementations without such support
	 */
	protected final StateImpl nfd;

	protected int kind;

	static {
		int result = ChannelNatives.initialize();
		if (result < 0) {
			throw new RuntimeException("Init Error: " + result);
		}
	}

	public Channel() {
		// XXX consider adding security check here, so only Classpath
		// code may create instances.
		this.nfd = new StateImpl();
		this.kind = Kind.OTHER;
	}

	/**
	 * This constructor is used by the POSIX reference implementation; other
	 * virtual machines need not support it.
	 *
	 * <strong>Important:</strong> do not call this in library code that is not
	 * specific to Classpath's reference implementation.
	 *
	 * @param native_fd
	 *            The native file descriptor integer.
	 * @throws IOException
	 */
	public Channel(final int native_fd) throws IOException {
		this();
		setNativeFD(native_fd);
	}

	public void setNativeFD(int fd) throws IOException {
		this.nfd.setNativeFD(fd);
	}

	@Override
	public void setBlocking(boolean blocking) throws IOException {
		int err = ChannelNatives.setBlocking(this.nfd.getNativeFD(), blocking, false);

		//retry native returned with result
		while(err == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			err = ChannelNatives.setBlocking(this.nfd.getNativeFD(), blocking, true);
		}

		if (err < 0) {
			throw new SocketException(Message.at(NetErrors.getInstance(), err));
		}
	}

	/**
	 * Throws an Exception if the given return code is an error code (i.e. <0).
	 */
	protected static void checkError(int ret) throws IOException {
		if (ret < 0) {
			switch (ret) {
			case NetErrors.J_EACCES:
			case NetErrors.J_ENETDOWN:
			case NetErrors.J_ENETUNREACH:
			case NetErrors.J_EINVAL:
			case NetErrors.J_EMSGSIZE:
			case NetErrors.J_ENOMEM:
			case NetErrors.J_EUNKNOWN:
				throw new IOException(Message.at(NetErrors.getInstance(), ret));
			case NetErrors.J_ETIMEDOUT:
				throw new SocketTimeoutException();
			case NetErrors.J_EHOSTUNKNOWN:
				throw new UnknownHostException();
			case NetErrors.J_NET_NATIVE_NOT_IMPLEMENTED:
				throw new RuntimeException(NetErrors.NOT_IMPLEMENTED);
			default:
				throw new SocketException(Message.at(NetErrors.getInstance(), ret));
			}

		}
	}

	/**
	 * Check if the socket is closed or if the connection is reset by the peer
	 *
	 * @param onlyOnClose
	 *            if true, check only if the socket is closed.
	 * @throws IOException
	 */
	protected void checkSockConnection(boolean onlyOnClose) throws IOException {
		// Nothing to do. VMChannel does not refer to a socket or server socket
		// Override by VMChannelNet
	}

	// Close.
	public static void close(int fd) throws IOException {
		int err = ChannelNatives.closeNative(fd, false);

		//retry native returned with result
		while(err == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			err = ChannelNatives.closeNative(fd, true);
		}

		checkError(err);
	}

	@Override
	public void close() throws IOException {
		this.nfd.close();
	}

	// Inner classes.

	/**
	 * A wrapper for a native file descriptor integer. This tracks the state of
	 * an open file descriptor, and ensures that
	 *
	 * This class need not be fully supported by virtual machines; if a virtual
	 * machine does not use integer file descriptors, or does and wishes to hide
	 * that, then the methods of this class may be stubbed out.
	 *
	 * System-specific classes that depend on access to native file descriptor
	 * integers SHOULD declare this fact.
	 */
	public final class StateImpl implements State {

		private int native_fd;
		private boolean valid;
		private boolean closed;

		StateImpl() {
			this.native_fd = -1;
			this.valid = false;
			this.closed = false;
		}

		@Override
		public boolean isValid() {
			return this.valid;
		}

		@Override
		public boolean isClosed() {
			synchronized (this) {
				return this.closed;
			}
		}

		@Override
		public int getNativeFD() throws IOException {
			if (!this.valid) {
				throw new IOException("invalid file descriptor");
			}
			return this.native_fd;
		}

		void setNativeFD(final int native_fd) throws IOException {
			if (this.valid) {
				throw new IOException("file descriptor already initialized");
			}
			this.native_fd = native_fd;
			this.valid = true;
		}

		@Override
		public void close() throws IOException {
			synchronized (this) {
				if (!this.valid) {
					throw new IOException("invalid file descriptor");
				}

				if (isSockChannel()) {
					// Before close socket, call shutdown to unblock the channel
					int err = ChannelNatives.shutdown(this.native_fd, false);

					//retry native returned with result
					while(err == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
						err = ChannelNatives.shutdown(this.native_fd, true);
					}

					checkError(err);
				}
				Channel.close(this.native_fd);
				this.valid = false;
				this.closed = true;
			}

		}

		@Override
		public String toString() {
			if (this.closed) {
				return "<<closed>>";
			}
			if (!this.valid) {
				return "<<invalid>>";
			}
			return String.valueOf(this.native_fd);
		}

	}

	/**
	 * Check if this channel refers to a Socket/ServerSocket
	 *
	 * @return
	 */
	protected boolean isSockChannel() {
		return (this.kind == Kind.SOCK_STREAM || this.kind == Kind.SOCK_DGRAM);
	}

	/**
	 * An enumeration of possible kinds of channel.
	 */
	static class Kind {

		/** A streaming (TCP) socket. */
		static final int SOCK_STREAM = 0;

		/** A datagram (UDP) socket. */
		static final int SOCK_DGRAM = 2;

		/** A file. */
		static final int FILE = 3;

		/** Something else; not a socket or file. */
		static final int OTHER = 4;
	}

	@Override
	public State getState() {
		return this.nfd;
	}

	@Override
	public void setOption(int optionId, Object optionValue)
			throws SocketException {
		try {
			int value;
			if (optionValue instanceof Integer) {
				value = ((Integer) optionValue).intValue();
			} else if (optionValue instanceof Boolean) {
				value = ((Boolean) optionValue).booleanValue() ? 1 : 0;
				// special handling for linger: -1 is a non-error value meaning "disabled"
				if ((value == 0) && (optionId == SocketOptions.SO_LINGER)) {
					value = -1;
				}
			} else if (optionValue instanceof java.net.NetworkInterface) {
				Enumeration<InetAddress> addrs = ((java.net.NetworkInterface)optionValue).getInetAddresses();
				while(addrs.hasMoreElements()) {
					setOption(optionId, addrs.nextElement());
				}
				// need to break from here, as the following block is for the int/boolean values
				return;
			} else if (optionValue instanceof InetAddress) {
				if (!(optionValue instanceof Inet4Address) && !(optionValue instanceof Inet6Address)) {
					throw new SocketException(NetErrors.UNSUPPORTED_INTERNET_ADDRESS);
				}
				byte[] addr = ((InetAddress)optionValue).getAddress();
				int ret = ChannelNatives.setOption(this.nfd.getNativeFD(),
						optionId, addr, addr.length, false);

				////retry write operation until data has been sent or an error occurred (or a timeout)
				while(ret == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
					ChannelNatives.setOption(this.nfd.getNativeFD(),
							optionId, addr, addr.length, true);
				}
				if (ret< 0) {
					throw new SocketException(Message.at(NetErrors.getInstance(), ret));
				}
				// need to break from here, as the following block is for the int/boolean values
				return;

			} else {
				throw new IllegalArgumentException("option value type " + optionValue.getClass().getName());
			}

			// if we reach here, then we have an int value
			int err = ChannelNatives.setOption(this.nfd.getNativeFD(), optionId, value, false);

			// retry native returned with result
			while (err == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				err = ChannelNatives.setOption(this.nfd.getNativeFD(), optionId, value, true);
			}

			if (err < 0) {
				throw new SocketException(Message.at(NetErrors.getInstance(), err));
			}
		} catch (IOException ioe) {
			SocketException se = new SocketException(/* CLDC-1.1-COMPLIANCE */ioe.getMessage());
			/*
			 * CLDC-1.1-COMPLIANCE se.initCause(ioe);
			 */
			throw se;
		}
	}

	/**
	 * Get a socket option. This implementation is only required to support
	 * socket options that are boolean values, which include:
	 *
	 * SocketOptions.IP_MULTICAST_LOOP SocketOptions.SO_BROADCAST
	 * SocketOptions.SO_KEEPALIVE SocketOptions.SO_OOBINLINE
	 * SocketOptions.SO_REUSEADDR SocketOptions.TCP_NODELAY
	 *
	 * and socket options that are integer values, which include:
	 *
	 * SocketOptions.IP_TOS SocketOptions.SO_LINGER SocketOptions.SO_RCVBUF
	 * SocketOptions.SO_SNDBUF SocketOptions.SO_TIMEOUT
	 *
	 * @param optionId
	 *            The option ID to fetch.
	 * @return A {@link Boolean} or {@link Integer} containing the socket
	 *         option.
	 * @throws SocketException
	 */
	@Override
	public Object getOption(int optionId) throws SocketException {
		switch (optionId) {
		case SocketOptions.IP_MULTICAST_IF:
		case SocketOptions.IP_MULTICAST_IF2:
			return getOptionValueByteArray(optionId);
		default:
			return getOptionValueInt(optionId);
		}
	}

	/**
	 * @param optionId
	 * @return
	 */
	private Object getOptionValueInt(int optionId) throws SocketException {
		int value;
		try {
			value = ChannelNatives.getOption(this.nfd.getNativeFD(), optionId, false);

			//retry native returned with result
			while(value == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
				value = ChannelNatives.getOption(this.nfd.getNativeFD(), optionId, true);
			}

			if (value < 0) {
				// special handling for linger: -1 is a non-error value meaning "disabled"
				if ((optionId != SocketOptions.SO_LINGER) || (value != -1)) {
					throw new SocketException(Message.at(NetErrors.getInstance(), value));
				}
			}
		} catch (IOException ioe) {
			SocketException se = new SocketException(
					/* CLDC-1.1-COMPLIANCE */ioe.getMessage());
			/*
			 * CLDC-1.1-COMPLIANCE se.initCause(ioe);
			 */
			throw se;
		}

		switch (optionId) {
		case SocketOptions.IP_MULTICAST_LOOP:
		case SocketOptions.SO_REUSEADDR:
		case SocketOptions.SO_OOBINLINE:
		case SocketOptions.SO_BROADCAST:
		case SocketOptions.SO_KEEPALIVE:
		case SocketOptions.TCP_NODELAY:
			/*
			 * CLDC-1.1-COMPLIANCE return Boolean.valueOf(value != 0);
			 */
			return (value != 0) ? Boolean.TRUE : Boolean.FALSE;

		case SocketOptions.IP_TOS:
		case SocketOptions.SO_RCVBUF:
		case SocketOptions.SO_SNDBUF:
		case SocketOptions.SO_TIMEOUT:
		case SocketOptions.SO_LINGER:
			return new Integer(value);

		default:
			throw new SocketException("getting option " + optionId
					+ " not supported here");
		}
	}

	private Object getOptionValueByteArray(int optionId) throws SocketException {
		try {
			byte[] optionBuffer = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
			int ret;

			ret = ChannelNatives.getOption(this.nfd.getNativeFD(), optionId, optionBuffer, optionBuffer.length,
					false);

			////retry write operation until data has been sent or an error occurred (or a timeout)
			while(ret == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
				ret = ChannelNatives.getOption(this.nfd.getNativeFD(), optionId, optionBuffer, optionBuffer.length,
						true);
			}

			if (ret < 0) {
				throw new SocketException(Message.at(NetErrors.getInstance(), ret));
			}
			byte[] addr = new byte[ret];
			System.arraycopy(optionBuffer, 0, addr, 0, addr.length);
			InetAddress inetAddr = InetAddress.getByAddress(addr);
			switch (optionId) {
			case SocketOptions.IP_MULTICAST_IF:
				return inetAddr;
			case SocketOptions.IP_MULTICAST_IF2:
				return java.net.NetworkInterface.getByInetAddress(inetAddr);
			default:
				throw new SocketException("unexpected option " + optionId);
			}

		} catch (IOException ioe) {
			SocketException se = new SocketException(
					/* CLDC-1.1-COMPLIANCE */ioe.getMessage());
			/*
			 * CLDC-1.1-COMPLIANCE se.initCause(ioe);
			 */
			throw se;
		}
	}
	/**
	 * Binds this socket to the given local address and port.
	 *
	 * @param address
	 *            The address to bind to; the InetAddress is either an IPv4 or
	 *            IPv6 address.
	 * @throws IOException
	 *             If binding fails; for example, if the port in the given
	 *             InetSocketAddress is privileged, and the current process has
	 *             insufficient privileges.
	 */
	@Override
	public void bind(InetSocketAddress address) throws IOException {
		InetAddress addr = address.getAddress();
		byte[] addrBuffer = addr.getAddress();
		byte[] addressAsBytes = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
		System.arraycopy(addrBuffer, 0, addressAsBytes, 0, addrBuffer.length);
		int len = addrBuffer.length;

		int err = 0;
		if (addr instanceof Inet4Address || addr instanceof Inet6Address) {
			err = ChannelNatives.bind(this.nfd.getNativeFD(), addressAsBytes,
					len, address.getPort(), false);

			//retry native returned with result
			while(err == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
				err = ChannelNatives.bind(this.nfd.getNativeFD(), addressAsBytes,
						len, address.getPort(), true);
			}
		} else {
			throw new SocketException("unsupported address type");
		}
		if (err < 0) {
			switch(err) {
			case NetErrors.J_EADDRINUSE:
			case NetErrors.J_EACCES:
				throw new BindException(Message.at(NetErrors.getInstance(), err));
			default:
				throw new SocketException(Message.at(NetErrors.getInstance(), err));
			}
		}

	}

	/**
	 * Listen on this socket for incoming connections.
	 *
	 * @param backlog
	 *            The backlog of connections.
	 * @throws IOException
	 *             If listening fails.
	 * @see gnu.java.nio.VMChannel#accept()
	 */
	@Override
	public void listen(int backlog) throws IOException {
		int err = ChannelNatives.listen(this.nfd.getNativeFD(), backlog, false);

		//retry native returned with result
		while(err == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			err = ChannelNatives.listen(this.nfd.getNativeFD(), backlog, true);
		}

		if (err < 0) {
			throw new SocketException(Message.at(NetErrors.getInstance(), err));
		}
	}

}