/*
 * Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2010-2021, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.is2t.support.net;

import java.io.IOException;
import java.net.Inet4Address;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NetworkInterface;
import java.net.PortUnreachableException;
import java.net.SocketAddress;
import java.net.SocketException;
import java.util.Enumeration;

import com.is2t.errors.NetErrors;
import com.is2t.support.net.natives.DatagramSocketChannelNatives;
import com.is2t.support.net.natives.MulticastSocketChannelNatives;
import com.is2t.support.net.util.IPAddressConstants;
import com.is2t.support.net.util.IntRef;

import ej.bon.ByteArray;

public class DatagramSocketChannel extends SocketChannel implements IDatagramSocketChannel {

	/**
	 * The maximum host data size 16 bytes max for the host address (required for
	 * IPv6) + 4 bytes for the host port
	 */
	private static final int HOST_DATA_SIZE = 20;

	public DatagramSocketChannel() throws IOException {
		super();
		this.kind = Kind.SOCK_DGRAM;
	}

	@Override
	public void initDatagramSocket(boolean isMulticast) throws IOException {
		if (isMulticast) {
			initMulticastSocket();
		} else {
			initSocket(false);
		}
	}

	@Override
	public int send(byte[] data, int offset, int length, InetSocketAddress dst) throws IOException{
		InetAddress addr = dst.getAddress();
		if (addr == null) {
			throw new NullPointerException();
		}

		// byte[] hostAddr = addr.getAddress();

		byte[] addrBuffer = addr.getAddress();
		byte[] hostAddr = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
		System.arraycopy(addrBuffer, 0, hostAddr, 0, addrBuffer.length);
		int addrLength = addrBuffer.length;

		int ret = 0;
		if (addr instanceof Inet4Address || addr instanceof Inet6Address) {

			int fd = this.nfd.getNativeFD();
			int dstPort = dst.getPort();

			ret = DatagramSocketChannelNatives.send(fd, data, offset, length, hostAddr, addrLength, dstPort, false);

			//// retry write operation until data has been sent or an error occurred (or a
			//// timeout)
			while (ret == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				ret = DatagramSocketChannelNatives.send(fd, data, offset, length, hostAddr, addrLength, dstPort,
						true);
			}

		} else {
			throw new SocketException(NetErrors.UNSUPPORTED_INTERNET_ADDRESS);
		}
		checkDatagramError(ret);
		return ret;
	}

	@Override
	public SocketAddress receive(byte[] data, int offset, int length,
			IntRef readLength) throws IOException {
		if (this.kind != Kind.SOCK_DGRAM) {
			throw new SocketException("not a datagram socket");
		}

		byte[] hostPort = new byte[HOST_DATA_SIZE];
		int fd = this.nfd.getNativeFD();
		int hostlen = 0;
		int receiveLength = 0;

		//The receive operation blocks until the fd become readable
		long ret = DatagramSocketChannelNatives.receive(fd, data, offset, length, hostPort, hostPort.length, false);

		//retry receive operation until there are available data or an error (or a timeout)
		while(ret == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			ret = DatagramSocketChannelNatives.receive(fd, data, offset, length, hostPort, hostPort.length, true);
		}

		// extract host length and data receive lenght from the long value returned by native
		hostlen = (int)(ret & 0xFFFFFFFFl);
		receiveLength = (int)((ret >> 32l) & 0xFFFFFFFFl);

		checkDatagramError(hostlen);

		// update the read value
		readLength.setValue(receiveLength);

		if (hostlen == 0) {
			return null;
		} else {
			byte[] addr = new byte[hostlen];
			int addrLength = addr.length;
			System.arraycopy(hostPort, 0, addr, 0, addrLength);
			int port = ByteArray.readInt(hostPort, addrLength);

			InetAddress inetAddress;
			if (hostlen == 4) { // IPv4
				inetAddress = Inet4Address.getByAddress(addr);
			} else if (hostlen == 16) { // IPv6
				inetAddress = Inet6Address.getByAddress(addr);
			} else {
				throw new SocketException(
						"host address received with invalid length: "
								+ hostlen);
			}

			return new InetSocketAddress(inetAddress, port);
		}
	}

	/**
	 * Overrides the default behavior of checkError with datagram specifics.
	 * Should be a virtual method actually overriding Channel.checkError but
	 * unfortunately this method is static and cannot be virtual.
	 *
	 * @param returnCode
	 *            the value that may contain an error code (i.e. <0)
	 */
	protected void checkDatagramError(int returnCode) throws IOException {
		if (returnCode < 0) {
			if (returnCode == NetErrors.J_ECONNREFUSED) {
				throw new PortUnreachableException();
			} else {
				checkError(returnCode);
			}
		}
	}

	/**
	 * Disconnect this channel, if it is a datagram socket. Disconnecting a
	 * datagram channel will disassociate it from any address, so the socket
	 * will remain open, but can send and receive datagrams from any address.
	 *
	 * @throws IOException
	 *             If disconnecting this channel fails, or if this channel is
	 *             not a datagram channel.
	 */
	@Override
	public void disconnect() throws IOException {
		int err = DatagramSocketChannelNatives.disconnect(this.nfd.getNativeFD(), false);

		//retry native returned with result
		while(err == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			err = DatagramSocketChannelNatives.disconnect(this.nfd.getNativeFD(), true);
		}

		checkError(err);
	}

	@Override
	public void setTimeToLive(int ttl) throws IOException {
		int err = MulticastSocketChannelNatives.setTimeToLive(this.nfd.getNativeFD(), ttl, false);

		//retry native returned with result
		while(err == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			err = MulticastSocketChannelNatives.setTimeToLive(this.nfd.getNativeFD(), ttl, true);
		}

		checkError(err);
	}

	@Override
	public int getTimeToLive() throws IOException {
		int ttl = MulticastSocketChannelNatives.getTimeToLive(this.nfd.getNativeFD(), false);

		//retry native returned with result
		while(ttl == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			ttl = MulticastSocketChannelNatives.getTimeToLive(this.nfd.getNativeFD(), true);
		}

		checkError(ttl);

		return ttl;
	}

	private int fillInterfaceAddress(byte[] immortal, NetworkInterface netIf) {
		InetAddress addr = null;
		if (netIf != null) {
			Enumeration<InetAddress> addresses = netIf.getInetAddresses();
			if (addresses.hasMoreElements()) {
				addr = addresses.nextElement();
			}
		}
		if (addr != null) {
			byte[] addrBytes = addr.getAddress();
			int len = addrBytes.length;
			System.arraycopy(addrBytes, 0, immortal, 0, len);
			return len;
		}
		return 0;
	}

	private void joinOrLeave(boolean join, InetAddress addr, NetworkInterface netIf) throws IOException {
		if (addr == null) {
			throw new NullPointerException();
		}

		byte[] mcastAddr = addr.getAddress();
		int mcastAddrLength = mcastAddr.length;
		byte[] netIfAddr = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
		int netIfAddrLength = fillInterfaceAddress(netIfAddr, netIf);

		int ret = 0;
		if(addr instanceof Inet4Address || addr instanceof Inet6Address) {

			int fd = this.nfd.getNativeFD();

			ret = MulticastSocketChannelNatives.joinOrLeave(fd, join, mcastAddr, mcastAddrLength, netIfAddr,
					netIfAddrLength, false);

			////retry join/leave operation until data has been sent or an error occurred (or a timeout)
			while(ret == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
				ret = MulticastSocketChannelNatives.joinOrLeave(fd, join, mcastAddr, mcastAddrLength,
						netIfAddr, netIfAddrLength, true);
			}
		} else {
			throw new SocketException(NetErrors.UNSUPPORTED_INTERNET_ADDRESS);
		}
		checkDatagramError(ret);

	}

	@Override
	public void join(InetAddress addr, NetworkInterface netIf) throws IOException {
		joinOrLeave(true, addr, netIf);
	}

	@Override
	public void leave(InetAddress addr, NetworkInterface netIf) throws IOException {
		joinOrLeave(false, addr, netIf);
	}
}
