/*
 * Copyright (C) 2017-2019, MicroEJ Corp. - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.security.cipher;

import java.security.InvalidKeyException;
import java.security.ProviderException;

import com.is2t.support.security.NativeResource;

import ej.bon.ByteArray;
import ej.sni.NativeException;

/**
 *
 */
public class NativeCipherSpi extends AbstractCipherSpi {

	/**
	 * @see #nativeGetTransformationDescription(byte[], byte[])
	 */
	public static final int CICPHER_DESCRIPTION_SIZE = ByteArray.INT_SIZE * 3;

	/**
	 * Native identifier for the cipher. Initialized during engine init.
	 */
	private int nativeId;

	/**
	 * Native resource related to the cipher. Initialized during engine init.
	 */
	private NativeCipherResource nativeResource;

	private final int nativeTransformationId;

	public NativeCipherSpi(int nativeTransformationId, int blockSize, int unitBytes,
			int cipherMode) {
		super(blockSize, unitBytes, cipherMode);
		this.nativeTransformationId = nativeTransformationId;
	}

	@Override
	protected int getBufferedLength() {
		return nativeGetBufferedLength(nativeTransformationId, nativeId);
	}

	@Override
	protected byte[] engineGetIV() {
		int ivLength = nativeGetIVLength(nativeTransformationId, nativeId);
		if (ivLength < 0) {
			return null;
		}

		byte[] iv = new byte[ivLength];
		nativeGetIV(nativeTransformationId, nativeId, iv, ivLength);
		return iv;
	}

	@Override
	protected void engineInit(boolean decrypting, String algorithm, byte[] key, byte[] iv) throws InvalidKeyException {
		// TODO check key algorithm string value

		try {

			if (nativeResource != null) {
				// close and reset the existing resource
				nativeResource.close();
				this.nativeId = -1;
				this.nativeResource = null;
			}

			int nativeTransformationId = this.nativeTransformationId;
			try {
				// Maybe some resources was garbage collected.
				// Try to release their related underlying native resources without trigger
				// the GC before.
				NativeResource.recycle(false);

				this.nativeId = nativeInit(nativeTransformationId, decrypting, key, key.length, iv, iv.length);
				// the initialization succeeds; create the associated resource.
				this.nativeResource = new NativeCipherResource(this, nativeTransformationId, this.nativeId);

			} catch (NativeException e) {
				// An error occurs during initialization, may be there is no space left to
				// allocate new native resource.
				// Trigger a new fresh GC and release new potential useless resources and try
				// the initialization again!
				NativeResource.recycle(true);
				this.nativeId = nativeInit(nativeTransformationId, decrypting, key, key.length, iv, iv.length);
				// the initialization succeeds; create the associated resource.
				this.nativeResource = new NativeCipherResource(this, nativeTransformationId, this.nativeId);
			}

		} catch (NativeException e) {
			throw new ProviderException(e);
		}
	}

	@Override
	protected int decrypt(byte[] buffer, int i, int len, byte[] output, int outputOffset) {
		return nativeDecrypt(nativeTransformationId, nativeId, buffer, i, len, output, outputOffset);
	}

	@Override
	protected int encrypt(byte[] buffer, int i, int len, byte[] output, int outputOffset) {
		return nativeEncrypt(nativeTransformationId, nativeId, buffer, i, len, output, outputOffset);
	}

	/**
	 * Gets for the given transformation the cipher description.
	 * <p>
	 * <code>description</code> must be filled-in with:
	 * <ul>
	 * <li>[0-3]: blockSize: the block size (in bytes), or 0 if the underlying
	 * algorithm is not a block cipher</li>
	 * <li>[4-7]: unitBytes: unit size (number of input bytes that can be processed
	 * at a time)</li>
	 * <li>[8-11]: cipherMode: one of the <code>*_MODE</code> values defined in
	 * {@link AbstractCipherSpi}</li>
	 * </ul>
	 *
	 * @param transformation
	 *            Null terminated string that describes the transformation
	 * @return transformation ID on success or -1 on error.
	 */
	native public static int nativeGetTransformationDescription(byte[] transformation, byte[] description);

	native private static int nativeGetBufferedLength(int nativeTransformationId, int nativeId);

	native private static void nativeGetIV(int nativeTransformationId, int nativeId, byte[] iv, int ivLength);

	/**
	 * Return -1 if no IV, otherwise returns the length of the IV.
	 */
	native private static int nativeGetIVLength(int nativeTransformationId, int nativeId);

	/**
	 * Throw NativeException on error. Returns a nativeId.
	 */
	native private static int nativeInit(int nativeTransformationId, boolean decrypting, byte[] key, int keyLength,
			byte[] iv, int ivLength) throws NativeException;

	native private static int nativeDecrypt(int nativeTransformationId, int nativeId, byte[] buffer, int i, int len,
			byte[] output, int outputOffset) throws NativeException;

	native private static int nativeEncrypt(int nativeTransformationId, int nativeId, byte[] buffer, int i, int len,
			byte[] output, int outputOffset) throws NativeException;

	native public static void nativeClose(int nativeTransformationId, int nativeId) throws NativeException;

}
