/*
 * Java
 *
 * Copyright 2019-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.verisilicon.vivante;

import java.util.HashMap;
import java.util.Map;

/**
 * Provides access to vg_lite equivalent native APIs specific for the matrix operations.
 */
public class Nvg_lite_matrix {
	private static Map<Integer, Matrix> matrixMap = new HashMap<>();

	/**
	 * Creates an identity matrix.
	 * <p>
	 *
	 * @return identity matrix.
	 */
	public static int create() {
		Matrix m = new Matrix();
		matrixMap.put(m.getHandle(), m);
		m.identity();

		return m.getHandle();
	}

	/**
	 * Destroy a matrix.
	 * <p>
	 *
	 * @param handle
	 *            matrix to destroy.
	 *
	 * @return 0.
	 */
	public static int destroy(int handle) {
		matrixMap.remove(handle);

		return 0;
	}

	/**
	 * Load the identity matrix.
	 * <p>
	 *
	 * @param handle
	 *            matrix on which to execute operation.
	 * @return 0
	 */
	public static int identity(int handle) {
		Matrix m = get(handle);

		if (m != null) {
			m.identity();
		}

		return 0;
	}

	/**
	 * Translate the matrix to a new position.
	 * <p>
	 *
	 * @param handle
	 *            matrix on which to execute operation.
	 * @param x
	 *            X location of the transformation.
	 * @param y
	 *            Y location of the transformation.
	 * @return 0
	 */
	public static int translate(int handle, float x, float y) {
		Matrix m = get(handle);

		if (m != null) {
			m.translate(x, y);
		}

		return 0;
	}

	/**
	 * Scale the matrix in both x and y directions.
	 * <p>
	 *
	 * @param handle
	 *            matrix on which to execute operation.
	 * @param x
	 *            Horizontal scale.
	 * @param y
	 *            Vertical scale.
	 * @return 0
	 */
	public static int scale(int handle, float x, float y) {
		Matrix m = get(handle);

		if (m != null) {
			m.scale(x, y);
		}

		return 0;
	}

	/**
	 * Rotate the matrix.
	 * <p>
	 *
	 * @param handle
	 *            matrix on which to execute operation.
	 * @param degrees
	 *            Number of degrees to rotate the matrix around. Positive numbers rotate counter clock wise.
	 * @return 0
	 */
	public static int rotate(int handle, float degrees) {
		Matrix m = get(handle);

		if (m != null) {
			m.rotate(degrees);
		}

		return 0;
	}

	/**
	 * Set perspective matrix.
	 * <p>
	 *
	 * @param handle
	 *            matrix on which to execute operation.
	 * @param px
	 *            w0 of perspective transformation matrix.
	 * @param py
	 *            w1 of perspective transformation matrix.
	 * @return 0
	 */
	public static int vg_lite_matrix_perspective(int handle, float px, float py) {
		throw new UnsupportedOperationException();
	}

	/**
	 * Get Matrix from handle
	 * <p>
	 *
	 * @param handle
	 *            matrix hanlde
	 * @return Matrix
	 */
	public static Matrix get(int handle) {
		return matrixMap.get(Integer.valueOf(handle));
	}

	/**
	 * Copy matrix parameters from one matrix to an other.
	 *
	 * @param srcHandle
	 *            Source matrix handle
	 * @param destHandle
	 *            Destination matrix handle
	 * @return 0
	 */
	public static int copy(int srcHandle, int destHandle) {
		Matrix srcMatrix = get(srcHandle);
		Matrix destMatrix = get(destHandle);

		if (srcMatrix != null && destMatrix != null) {
			destMatrix.copy(srcMatrix);
		}

		return 0;
	}
}
