/*
 * Java
 *
 * Copyright 2016-2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package ej.ecom.network;

import java.net.InetAddress;

/**
 * An IP configuration describes the settings of an interface.
 *
 * An instance may be sparse. <code>null</code> fields indicate properties that have not been set (e.g. DNS may be null
 * when statically set, meaning that there is no known DNS server) or that are not known yet (e.g. DHCP is used but we
 * have been assigned an IP address yet).
 */
public class IPConfiguration {

	private boolean isStatic;
	private InetAddress ip;
	private InetAddress netmask;
	private InetAddress gateway;

	private boolean isDNSStatic;
	private InetAddress[] dns;

	public boolean isStatic() {
		return isStatic;
	}

	/**
	 * Force the use of DHCP or not.
	 *
	 * @param use
	 *            <code>true</code> to use DHCP, <code>false</code> for a static configuration
	 */
	public void useDHCP(boolean use) {
		isStatic = !use;
	}

	public InetAddress getIP() {
		return ip;
	}

	/**
	 * Set the IP address.
	 *
	 * @param ip
	 *            the IP address
	 */
	public void setIP(InetAddress ip) {
		this.ip = ip;
	}

	public InetAddress getNetmask() {
		return netmask;
	}

	/**
	 * Set the netmask.
	 *
	 * @param netmask
	 *            the netmask
	 */
	public void setNetmask(InetAddress netmask) {
		this.netmask = netmask;
	}

	public InetAddress getGateway() {
		return gateway;
	}

	/**
	 * Set the IP address of the gateway.
	 *
	 * @param gateway
	 *            the IP address of the gateway
	 */
	public void setGateway(InetAddress gateway) {
		this.gateway = gateway;
	}

	public boolean isDNSStatic() {
		return isDNSStatic;
	}

	/**
	 * When DHCP is used, it is possible to have statically configured DNS. This method allows to select between DNS given by DHCP server or set with
	 * {@link #setDNS(InetAddress[])}.
	 *
	 * @param use
	 */
	public void useStaticDNS(boolean use) {
		isDNSStatic = use;
	}

	public InetAddress[] getDNS() {
		return dns;
	}

	/**
	 * Set the addresses of the DNS servers.
	 *
	 * @param DNS
	 *            the addresses of the DNS servers
	 */
	public void setDNS(InetAddress[] DNS) {
		this.dns = DNS;
	}

	@Override
	public String toString() {
		String str = "IPConfigurationImpl [isStatic=" + isStatic + ", ip=" + ip + ", netmask=" + netmask + ", gateway=" + gateway + ", isDNSStatic="
				+ isDNSStatic + ", dns=";

		if ((dns == null) || dns.length == 0) {
			str += "null";
		} else {
			for (int i = 0; i < dns.length - 1; i++) {
				str += dns[i];
				str += ", ";
			}
			str += dns[dns.length - 1];
		}

		str += "]";
		return str;
	}

}
