/*
 * Copyright (c) 1995, 2013, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2015-2016, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package java.net;

import java.io.IOException;

import com.is2t.support.DependencyInjectionHelper;
import com.is2t.support.net.IDatagramSocketImplFactory;
import com.is2t.support.net.security.INetSecurityManager;
import com.is2t.support.net.security.INetSecurityManagerFactory;

public
class DatagramSocket implements java.io.Closeable {

	private static INetSecurityManagerFactory NetSecurityManagerFactory = (INetSecurityManagerFactory) DependencyInjectionHelper.getImplementation(INetSecurityManagerFactory.class);
	private static INetSecurityManager NetSecurityManager = NetSecurityManagerFactory.getNetSecurityManager();
	private static IDatagramSocketImplFactory DatagramSocketImplFactory = (IDatagramSocketImplFactory) DependencyInjectionHelper.getImplementation(IDatagramSocketImplFactory.class);

	/**
	 * Various states of this socket.
	 */
	private boolean created = false;
	private boolean bound = false;
	private boolean closed = false;
	private final Object closeLock = new Object();

	/*
	 * The implementation of this DatagramSocket.
	 */
	DatagramSocketImpl impl;


	/*
	 * Connection state:
	 * ST_NOT_CONNECTED = socket not connected
	 * ST_CONNECTED = socket connected
	 * ST_CONNECTED_NO_IMPL = socket connected but not at impl level
	 */
	static final int ST_NOT_CONNECTED = 0;
	static final int ST_CONNECTED = 1;
	static final int ST_CONNECTED_NO_IMPL = 2;

	int connectState = ST_NOT_CONNECTED;

	/*
	 * Connected address & port
	 */
	InetAddress connectedAddress = null;
	int connectedPort = -1;

	/**
	 * Connects this socket to a remote socket address (IP address + port number).
	 * Binds socket if not already bound.
	 * <p>
	 * @param   address The remote address.
	 * @param   port    The remote port
	 * @throws  SocketException if binding the socket fails.
	 */
	private synchronized void connectInternal(InetAddress address, int port) throws SocketException {
		if (port < 0 || port > 0xFFFF) {
			throw new IllegalArgumentException("connect: " + port);
		}
		if (address == null) {
			throw new IllegalArgumentException("connect: null address");
		}
		checkAddress (address, "connect");
		if (isClosed()) {
			return;
		}
		SecurityManager security = System.getSecurityManager();
		if (security != null) {

			if (address.isMulticastAddress()) {
				NetSecurityManager.checkMulticast(security, address);
			} else {
				NetSecurityManager.checkConnect(security, address.getHostAddress(), port);
				NetSecurityManager.checkAccept(security, address.getHostAddress(), port);
			}
		}

		if (!isBound()) {
			bind(new InetSocketAddress(0));
		}

		try {
			getImpl().connect(address, port);

			// socket is now connected by the impl
			connectState = ST_CONNECTED;
			connectedAddress = address;
			connectedPort = port;
		} catch (SocketException se) {

			connectState = ST_NOT_CONNECTED;
		}

	}


	public DatagramSocket() throws SocketException {
		this(new InetSocketAddress(0));
	}

	/**
	 * Creates an unbound datagram socket with the specified
	 * DatagramSocketImpl.
	 *
	 * @param impl an instance of a <B>DatagramSocketImpl</B>
	 *        the subclass wishes to use on the DatagramSocket.
	 * @since   1.4
	 */
	protected DatagramSocket(DatagramSocketImpl impl) {
		if (impl == null) {
			throw new NullPointerException();
		}
		this.impl = impl;
	}

	public DatagramSocket(SocketAddress bindaddr) throws SocketException {
		// create a datagram socket.
		createImpl();
		if (bindaddr != null) {
			try {
				bind(bindaddr);
			} finally {
				if (!isBound()) {
					close();
				}
			}
		}
	}

	public DatagramSocket(int port) throws SocketException {
		this(port, null);
	}

	public DatagramSocket(int port, InetAddress laddr) throws SocketException {
		this(new InetSocketAddress(laddr, port));
	}


	void createImpl() throws SocketException {
		if (impl == null) {
			if (factory != null) {
				impl = factory.createDatagramSocketImpl();
			} else {
				boolean isMulticast = (this instanceof MulticastSocket) ? true : false;
				impl = DatagramSocketImplFactory.createDatagramSocketImpl(isMulticast);
			}
		}
		// creates a udp socket
		impl.create();
		created = true;
	}

	/**
	 * Get the {@code DatagramSocketImpl} attached to this socket,
	 * creating it if necessary.
	 *
	 * @return  the {@code DatagramSocketImpl} attached to that
	 *          DatagramSocket
	 * @throws SocketException if creation fails.
	 * @since 1.4
	 */
	public DatagramSocketImpl getImpl() throws SocketException {
		if (!created) {
			createImpl();
		}
		return impl;
	}

	public synchronized void bind(SocketAddress addr) throws SocketException {
		if (isClosed()) {
			throw new SocketException("Socket is closed");
		}
		if (isBound()) {
			throw new SocketException("already bound");
		}
		if (addr == null) {
			addr = new InetSocketAddress(0);
		}
		if (!(addr instanceof InetSocketAddress)) {
			throw new IllegalArgumentException("Unsupported address type!");
		}
		InetSocketAddress epoint = (InetSocketAddress) addr;
		if (epoint.isUnresolved()) {
			throw new SocketException("Unresolved address");
		}
		InetAddress iaddr = epoint.getAddress();
		int port = epoint.getPort();
		checkAddress(iaddr, "bind");
		SecurityManager sec = System.getSecurityManager();
		if (sec != null) {
			NetSecurityManager.checkListen(sec, port);
		}
		try {
			getImpl().bind(port, iaddr);
		} catch (SocketException e) {
			getImpl().close();
			throw e;
		}
		bound = true;
	}

	void checkAddress (InetAddress addr, String op) {
		if (addr == null) {
			return;
		}
		if (!(addr instanceof Inet4Address || addr instanceof Inet6Address)) {
			throw new IllegalArgumentException(op + ": invalid address type");
		}
	}

	public void connect(InetAddress address, int port) {
		try {
			connectInternal(address, port);
		} catch (SocketException se) {
			throw new Error("connect failed", se);
		}
	}

	public void connect(SocketAddress addr) throws SocketException {
		if (addr == null) {
			throw new IllegalArgumentException("Address can't be null");
		}
		if (!(addr instanceof InetSocketAddress)) {
			throw new IllegalArgumentException("Unsupported address type");
		}
		InetSocketAddress epoint = (InetSocketAddress) addr;
		if (epoint.isUnresolved()) {
			throw new SocketException("Unresolved address");
		}
		connectInternal(epoint.getAddress(), epoint.getPort());
	}

	public void disconnect() {
		synchronized (this) {
			if (isClosed()) {
				return;
			}
			if (connectState == ST_CONNECTED) {
				impl.disconnect ();
			}
			connectedAddress = null;
			connectedPort = -1;
			connectState = ST_NOT_CONNECTED;
		}
	}

	public boolean isBound() {
		return bound;
	}

	public boolean isConnected() {
		return connectState != ST_NOT_CONNECTED;
	}

	public InetAddress getInetAddress() {
		return connectedAddress;
	}

	public int getPort() {
		return connectedPort;
	}

	public SocketAddress getRemoteSocketAddress() {
		if (!isConnected()) {
			return null;
		}
		return new InetSocketAddress(getInetAddress(), getPort());
	}

	public SocketAddress getLocalSocketAddress() {
		if (isClosed()) {
			return null;
		}
		if (!isBound()) {
			return null;
		}
		return new InetSocketAddress(getLocalAddress(), getLocalPort());
	}

	public void send(DatagramPacket p) throws IOException  {
		InetAddress packetAddress = null;
		synchronized (p) {
			if (isClosed()) {
				throw new SocketException("Socket is closed");
			}
			checkAddress (p.getAddress(), "send");
			if (connectState == ST_NOT_CONNECTED) {
				// check the address is ok wiht the security manager on every send.
				SecurityManager security = System.getSecurityManager();

				// The reason you want to synchronize on datagram packet
				// is because you don't want an applet to change the address
				// while you are trying to send the packet for example
				// after the security check but before the send.
				if (security != null) {
					if (p.getAddress().isMulticastAddress()) {
						NetSecurityManager.checkMulticast(security, p.getAddress());
					} else {
						NetSecurityManager.checkConnect(security, p.getAddress().getHostAddress(), p.getPort());
					}
				}
			} else {
				// we're connected
				packetAddress = p.getAddress();
				if (packetAddress == null) {
					p.setAddress(connectedAddress);
					p.setPort(connectedPort);
				} else if ((!packetAddress.equals(connectedAddress)) ||
						p.getPort() != connectedPort) {
					throw new IllegalArgumentException("connected address " +
							"and packet address" +
							" differ");
				}
			}
			// Check whether the socket is bound
			if (!isBound()) {
				bind(new InetSocketAddress(0));
			}
			// call the  method to send
			getImpl().send(p);
		}
	}

	public synchronized void receive(DatagramPacket p) throws IOException {
		synchronized (p) {
			if (!isBound()) {
				bind(new InetSocketAddress(0));
			}

			getImpl().receive(p);
			SecurityManager security = System.getSecurityManager();
			if(security != null && (connectState == ST_NOT_CONNECTED)){
				NetSecurityManager.checkAccept(security, p.getAddress().getHostAddress(), p.getPort());
			}
		}
	}

	public InetAddress getLocalAddress() {
		if (isClosed()) {
			return null;
		}
		InetAddress in = null;
		try {
			in = (InetAddress) getImpl().getOption(SocketOptions.SO_BINDADDR);
			if (in.isAnyLocalAddress()) {
				in = InetAddress.anyLocalAddress();
			}
			SecurityManager s = System.getSecurityManager();
			if (s != null) {
				NetSecurityManager.checkConnect(s, in.getHostAddress(), -1);
			}
		} catch (Exception e) {
			in = InetAddress.anyLocalAddress(); // "0.0.0.0"
		}
		return in;
	}

	public int getLocalPort() {
		if (isClosed()) {
			return -1;
		}
		try {
			return getImpl().getLocalPort();
		} catch (Exception e) {
			return 0;
		}
	}

	public synchronized void setSoTimeout(int timeout) throws SocketException {
		if (isClosed()) {
			throw new SocketException("Socket is closed");
		}
		getImpl().setOption(SocketOptions.SO_TIMEOUT, new Integer(timeout));
	}

	public synchronized int getSoTimeout() throws SocketException {
		if (isClosed()) {
			throw new SocketException("Socket is closed");
		}
		if (getImpl() == null) {
			return 0;
		}
		Object o = getImpl().getOption(SocketOptions.SO_TIMEOUT);
		/* extra type safety */
		if (o instanceof Integer) {
			return ((Integer) o).intValue();
		} else {
			return 0;
		}
	}

	public synchronized void setSendBufferSize(int size)
			throws SocketException{
		if (!(size > 0)) {
			throw new IllegalArgumentException("negative send size");
		}
		if (isClosed()) {
			throw new SocketException("Socket is closed");
		}
		getImpl().setOption(SocketOptions.SO_SNDBUF, new Integer(size));
	}

	public synchronized int getSendBufferSize() throws SocketException {
		if (isClosed()) {
			throw new SocketException("Socket is closed");
		}
		int result = 0;
		Object o = getImpl().getOption(SocketOptions.SO_SNDBUF);
		if (o instanceof Integer) {
			result = ((Integer)o).intValue();
		}
		return result;
	}

	public synchronized void setReceiveBufferSize(int size)
			throws SocketException{
		if (size <= 0) {
			throw new IllegalArgumentException("invalid receive size");
		}
		if (isClosed()) {
			throw new SocketException("Socket is closed");
		}
		getImpl().setOption(SocketOptions.SO_RCVBUF, new Integer(size));
	}

	public synchronized int getReceiveBufferSize()
			throws SocketException{
		if (isClosed()) {
			throw new SocketException("Socket is closed");
		}
		int result = 0;
		Object o = getImpl().getOption(SocketOptions.SO_RCVBUF);
		if (o instanceof Integer) {
			result = ((Integer)o).intValue();
		}
		return result;
	}

	public synchronized void setReuseAddress(boolean on) throws SocketException {
		if (isClosed()) {
			throw new SocketException("Socket is closed");
		}
		getImpl().setOption(SocketOptions.SO_REUSEADDR, Boolean.valueOf(on));
	}

	public synchronized boolean getReuseAddress() throws SocketException {
		if (isClosed()) {
			throw new SocketException("Socket is closed");
		}
		return ((Boolean)(getImpl().getOption(SocketOptions.SO_REUSEADDR))).booleanValue();
	}

	public synchronized void setBroadcast(boolean on) throws SocketException {
		if (isClosed()) {
			throw new SocketException("Socket is closed");
		}
		getImpl().setOption(SocketOptions.SO_BROADCAST, Boolean.valueOf(on));
	}

	public synchronized boolean getBroadcast() throws SocketException {
		if (isClosed()) {
			throw new SocketException("Socket is closed");
		}
		return ((Boolean)(getImpl().getOption(SocketOptions.SO_BROADCAST))).booleanValue();
	}

	public synchronized void setTrafficClass(int tc) throws SocketException {
		if (tc < 0 || tc > 255) {
			throw new IllegalArgumentException("tc is not in range 0 -- 255");
		}

		if (isClosed()) {
			throw new SocketException("Socket is closed");
		}
		getImpl().setOption(SocketOptions.IP_TOS, new Integer(tc));
	}

	public synchronized int getTrafficClass() throws SocketException {
		if (isClosed()) {
			throw new SocketException("Socket is closed");
		}
		return ((Integer)(getImpl().getOption(SocketOptions.IP_TOS))).intValue();
	}

	@Override
	public void close() {
		synchronized(closeLock) {
			if (isClosed()) {
				return;
			}
			impl.close();
			closed = true;
		}
	}

	public boolean isClosed() {
		synchronized(closeLock) {
			return closed;
		}
	}


	/**
	 * User defined factory for all datagram sockets.
	 */
	static DatagramSocketImplFactory factory;

	public static synchronized void
	setDatagramSocketImplFactory(DatagramSocketImplFactory fac)
			throws IOException
			{
		if (factory != null) {
			throw new SocketException("factory already defined");
		}
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			NetSecurityManager.checkSetFactory(security);
		}
		factory = fac;
			}
}
